#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from wield.bunch import Bunch

from .base import (
    ArgumentError,
    mapcheck_bool,
    mapcheck_positive_int,
    mapcheck_positive_int_orNone,
)


def mc_log_level(aid, aname, val):
    val = int(val)
    if val < 0 or val > 10:
        raise ArgumentError(
            "log_level values must be between 0-10, 0 being least verbose, 10 most."
        )
    return val


def hint_log_level(**kwargs):
    kw = dict(
        mapcheck=mc_log_level,
    )
    kw.update(**kwargs)
    return kw


kw_hints = Bunch(
    logging_module_use=dict(
        APignore=True,
        mapcheck=mapcheck_bool,
        about="Use the python logging module instead of print",
        aliases=["use_logging_module"],
    ),
    tee_logfile=dict(
        APgroup="logging",
        APpriority=10,
        default=None,
        about="""
        Filename to tee log output into, along with stdout/stderr. Will be overwritten.
        """,
    ),
    log_level=hint_log_level(
        APshort="-l",
        APpriority=20,
        APgroup="logging",
        mapcheck=mapcheck_positive_int,
        default=5,
        about="Log level default for all logging types",
        aliases=["verbosity"],
    ),
    log_level_alert=hint_log_level(
        APgroup="logging",
        APpriority=25,
        mapcheck=mapcheck_positive_int_orNone,
        about="""
        Data on useful statistical tests, particularly those which cause input
        data/SNR/emphasis to be reinterpreted. Typically logs at level 2-4.
        """,
        aliases=["verbosity_alert"],
    ),
    log_level_info=hint_log_level(
        APgroup="logging",
        APpriority=25,
        mapcheck=mapcheck_positive_int_orNone,
        about="""
        Logging on miscellaneous details.
        """,
        aliases=["verbosity_info"],
    ),
    log_level_debug=hint_log_level(
        APgroup="logging",
        APpriority=25,
        mapcheck=mapcheck_positive_int_orNone,
        default=0,
        about="""
        Debugging reports used for development.
        """,
        aliases=["verbosity_debug"],
    ),
    log_level_warn=hint_log_level(
        APgroup="logging",
        APpriority=25,
        mapcheck=mapcheck_positive_int_orNone,
        default=5,
        about="""
        Warnings about failed tests, or data operating in a regime unexpected to
        work or that validation should be made.
        """,
        aliases=["verbosity_warn"],
    ),
    log_level_rationale=hint_log_level(
        APgroup="logging",
        APpriority=25,
        mapcheck=mapcheck_positive_int_orNone,
        default=0,
        about="""
        Detailed explanations of tests and algorithms. Extremely verbose and
        intended for first users and digest reports.
        """,
        aliases=["verbosity_rationale"],
    ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from .residues import (
    ZPK2residues,
    ZPK2residues_scipy,
    residues2ZPK,
)

from .zpktf import (
    ZPKTF,
    asZPKTF,
    asMRRB,
)

from .zpk_with_data import (
    ZPKwData,
)

from .root_bunch import (
    RootBunch,
    RBAlgorithms,
    root_constraints,
    RootConstraints,
    RBalgo,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Calculation of the RMS of a filter with unit white noise passing through a ZPK.
This is done using an integral and residue calculus. The filter must have more
P1 than Z, unless the ZPK is in the Z domain,
where there is a natural cutoff frequency.
"""

import numpy as np

import numbers

from wield.bunch.depbunch import (
    DepBunch,
    depB_property,
    # NOARG,
)

from .root_bunch import (
    RBAlgorithms,
    RootBunch,
    root_constraints,
)

from .ratmath import (
    ZPKsum,
    ZPKprod,
    ZPKscalarprod,
    ZPKdiv,
    ZPKscalardiv,
    ZPKscalarsum,
    ZPKdivscalar,
)


class ZPKTF(DepBunch):
    RBalgo = RBAlgorithms()
    root_constraint = RBalgo.root_constraints.mirror_real

    def __build__(
        self,
        ZPK=None,
        zeros=None,
        poles=None,
        gain=None,
        F_nyquist_Hz="unknown",
        **kwargs
    ):
        super(ZPKTF, self).__build__(**kwargs)
        if zeros is None and ZPK is not None:
            try:
                zeros = ZPK.zeros
            except AttributeError:
                zeros = ZPK[0]
        if zeros is None:
            zeros = ()
        if poles is None and ZPK is not None:
            try:
                poles = ZPK.poles
            except AttributeError:
                poles = ZPK[1]
        if poles is None:
            poles = ()
        if gain is None and ZPK is not None:
            try:
                gain = ZPK.gain
            except AttributeError:
                gain = ZPK[2]
        if gain is None:
            gain = 1
        self.zeros = zeros
        self.poles = poles
        self.gain = gain
        self.F_nyquist_Hz = F_nyquist_Hz

    def __iter__(self):
        yield self.zeros.fullplane
        yield self.poles.fullplane
        yield self.gain

    def __str__(self):
        return "ZPKTF(Z={},P={},K={})".format(
            str(self.zeros), str(self.poles), str(self.gain)
        )

    @depB_property
    def test(self, val=1):
        return val

    @depB_property
    def gain(self, val):
        return val

    @depB_property
    def zeros(self, val):
        val = self.RBalgo.expect_atleast(val, constraint=self.root_constraint)
        return val

    @depB_property
    def poles(self, val):
        val = self.RBalgo.expect_atleast(val, constraint=self.root_constraint)
        return val

    @depB_property
    def order(self):
        return max(
            len(self.poles),
            len(self.zeros),
        )

    @depB_property
    def order_sos(self):
        return (
            max(
                len(self.poles),
                len(self.zeros),
            )
            + 1
        ) // 2

    @depB_property
    def order_relative(self):
        return len(self.zeros) - len(self.poles)

    @depB_property
    def order_total(self):
        return len(self.poles) + len(self.zeros)

    def xfer_eval(self, F_Hz):
        # TODO must add pole-zero rephasing for Z filters

        # TODO, make this name consistent in all classes
        if self.F_nyquist_Hz is None:
            X_grid = 1j * F_Hz
        else:
            # use Z^-1
            X_grid = np.exp(1j * np.pi * F_Hz / self.F_nyquist_Hz)
        h, lnG = self.poles.val_lnG(X_grid)
        h, lnG = self.zeros.val_lnG(X_grid, h=1 / h, lnG=-lnG)
        return h * (np.exp(lnG) * self.gain)

    def __add__(self, other):
        if isinstance(other, numbers.Real):
            Z3, P3, K3 = ZPKscalarsum(self, other)
            F_nyquist_Hz = self.F_nyquist_Hz
        elif isinstance(other, tuple):
            F_nyquist_Hz = common_nyquist(self, other)
            Z3, P3, K3 = ZPKsum(self, other)
        else:
            return NotImplemented
        return self.__class__(Z3, P3, K3, F_nyquist_Hz=F_nyquist_Hz)

    def __radd__(self, other):
        if isinstance(other, numbers.Real):
            Z3, P3, K3 = ZPKscalarsum(self, other)
            F_nyquist_Hz = self.F_nyquist_Hz
        elif isinstance(other, tuple):
            F_nyquist_Hz = common_nyquist(other, self)
            Z3, P3, K3 = ZPKsum(self, other)
        else:
            return NotImplemented
        return self.__class__(Z3, P3, K3, F_nyquist_Hz=F_nyquist_Hz)

    def __sub__(self, other):
        if isinstance(other, numbers.Real):
            Z3, P3, K3 = ZPKscalarsum(self, -other)
            F_nyquist_Hz = self.F_nyquist_Hz
        elif isinstance(other, tuple):
            F_nyquist_Hz = common_nyquist(self, other)
            Z2, P2, K2 = other
            Z3, P3, K3 = ZPKsum(self, (Z2, P2, -K2))
        else:
            return NotImplemented
        return self.__class__(Z3, P3, K3, F_nyquist_Hz=F_nyquist_Hz)

    def __rsub__(self, other):
        if isinstance(other, numbers.Real):
            Z3, P3, K3 = ZPKscalarsum(self, -other)
            F_nyquist_Hz = self.F_nyquist_Hz
        elif isinstance(other, tuple):
            F_nyquist_Hz = common_nyquist(other, self)
            Z2, P2, K2 = other
            Z3, P3, K3 = ZPKsum(self, (Z2, P2, -K2))
        else:
            return NotImplemented
        return self.__class__(Z3, P3, -K3, F_nyquist_Hz=F_nyquist_Hz)

    def __mul__(self, other):
        if isinstance(other, numbers.Real):
            Z3, P3, K3 = ZPKscalarprod(self, other)
            F_nyquist_Hz = self.F_nyquist_Hz
        elif isinstance(other, (tuple, ZPKTF)):
            F_nyquist_Hz = common_nyquist(self, other)
            Z3, P3, K3 = ZPKprod(self, other)
        else:
            return NotImplemented
        return self.__class__(zeros=Z3, poles=P3, gain=K3, F_nyquist_Hz=F_nyquist_Hz)

    def __rmul__(self, other):
        if isinstance(other, numbers.Real):
            Z3, P3, K3 = ZPKscalarprod(self, other)
            F_nyquist_Hz = self.F_nyquist_Hz
        elif isinstance(other, (tuple, ZPKTF)):
            F_nyquist_Hz = common_nyquist(other, self)
            Z3, P3, K3 = ZPKprod(other, self)
        else:
            return NotImplemented
        return self.__class__(zeros=Z3, poles=P3, gain=K3, F_nyquist_Hz=F_nyquist_Hz)

    def __truediv__(self, other):
        if isinstance(other, numbers.Real):
            Z3, P3, K3 = ZPKscalardiv(self, other)
            F_nyquist_Hz = self.F_nyquist_Hz
        elif isinstance(other, (tuple, ZPKTF)):
            F_nyquist_Hz = common_nyquist(self, other)
            Z3, P3, K3 = ZPKdiv(self, other)
        else:
            return NotImplemented
        return self.__class__(zeros=Z3, poles=P3, gain=K3, F_nyquist_Hz=F_nyquist_Hz)

    def __rtruediv__(self, other):
        if isinstance(other, numbers.Real):
            Z3, P3, K3 = ZPKdivscalar(other, self)
            F_nyquist_Hz = self.F_nyquist_Hz
        elif isinstance(other, (tuple, ZPKTF)):
            F_nyquist_Hz = common_nyquist(other, self)
            Z3, P3, K3 = ZPKdiv(other, self)
        else:
            return NotImplemented
        return self.__class__(zeros=Z3, poles=P3, gain=K3, F_nyquist_Hz=F_nyquist_Hz)

    # TODO, version check here?
    __div__ = __truediv__
    __rdiv__ = __rtruediv__

    def __pow__(self, other):
        if isinstance(other, numbers.Complex):
            if other.imag != 0:
                return NotImplemented
            other = other.real

        if not isinstance(other, numbers.Integral):
            residual = other % 1
            if residual < -1e-14 or residual > 1e-14:
                return NotImplemented
            other = int((other + 1e-14) // 1)

        if other == 0:
            return self.__class__((), (), 1)
        elif other == 1:
            return self
        elif other == -1:
            return self.__class__(
                self.poles, self.zeros, 1 / self.gain, F_nyquist_Hz=self.F_nyquist_Hz
            )
        elif other > 1:
            return self.__class__(
                tuple(self.zeros) * other,
                tuple(self.poles) * other,
                self.gain ** other,
                F_nyquist_Hz=self.F_nyquist_Hz,
            )
        elif other < 1:
            other = -other
            return self.__class__(
                tuple(self.poles) * other,
                tuple(self.zeros) * other,
                self.gain ** (-other),
                F_nyquist_Hz=self.F_nyquist_Hz,
            )

    def __neg__(self):
        return self.__class__(
            self.poles, self.zeros, -self.gain, F_nyquist_Hz=self.F_nyquist_Hz
        )

    def __pos__(self):
        return self

    def abs_sq(self, F_nyquist_Hz="unknown"):
        if self.F_nyquist_Hz != "unknown":
            F_nyquist_Hz = self.F_nyquist_Hz
        if F_nyquist_Hz == "unknown":
            raise RuntimeError(
                " F_nyquist_Hz must be known to perform the correct"
                " ZPK manipulation for abs_sq"
            )
        elif F_nyquist_Hz is None:
            return self.__class__(
                tuple(self.poles) + tuple(np.asarray(self.poles).conjugate()),
                tuple(self.zeros) + tuple(np.asarray(self.zeros).conjugate()),
                self.gain ** 2,
                F_nyquist_Hz=self.F_nyquist_Hz,
            )
        else:
            return self.__class__(
                tuple(self.poles) + tuple(1 / np.asarray(self.poles)),
                tuple(self.zeros) + tuple(1 / np.asarray(self.zeros)),
                self.gain ** 2,
                F_nyquist_Hz=self.F_nyquist_Hz,
            )

    def assert_F_nyquist_Hz(self, F_nyquist_Hz):
        if self.F_nyquist_Hz == "unknown":
            return
        if self.F_nyquist_Hz != F_nyquist_Hz:
            raise RuntimeError("Incompatible ZPK representations!")
        return F_nyquist_Hz


def common_nyquist(LHS, RHS):
    """
    Assumes LHS and RHS are a ZPK or has a
    F_nyquist_Hz property.
    """
    if isinstance(LHS, ZPKTF):
        LHS_F_nyquist_Hz = LHS.F_nyquist_Hz
    else:
        LHS_F_nyquist_Hz = "unknown"

    if isinstance(RHS, ZPKTF):
        RHS_F_nyquist_Hz = RHS.F_nyquist_Hz
    else:
        RHS_F_nyquist_Hz = "unknown"

    if LHS_F_nyquist_Hz == "unknown":
        return RHS_F_nyquist_Hz
    elif RHS_F_nyquist_Hz == "unknown":
        return LHS_F_nyquist_Hz
    if LHS_F_nyquist_Hz != RHS_F_nyquist_Hz:
        raise RuntimeError(
            (
                "Math Operation should not be done on ZPKs arising from "
                "different representations, LHS is {}, RHS is {}"
            ).format(LHS_F_nyquist_Hz, RHS_F_nyquist_Hz)
        )
    return LHS_F_nyquist_Hz


# TODO,
def asZPKTF(
    ZPK,
    complete=False,
    F_nyquist_Hz=None,
    delay_s=None,
):
    from .zpk_with_data import ZPKwData

    if isinstance(ZPK, ZPKTF):
        return ZPK
    elif isinstance(ZPK, (tuple, list)):
        Z, P, K = ZPK
        return ZPKTF(
            zeros=asMRRB(Z, complete=complete),
            poles=asMRRB(P, complete=complete),
            gain=K,
            F_nyquist_Hz=F_nyquist_Hz,
        )
    elif isinstance(ZPK, ZPKwData):
        return ZPKTF(
            zeros=ZPK.zeros * ZPK.zeros_overlay,
            poles=ZPK.poles * ZPK.poles_overlay,
            gain=ZPK.gain,
            F_nyquist_Hz=F_nyquist_Hz,
        )
    # last ditch effort if it is some other wield.iirrational type
    return asZPKTF(ZPK.ZPKrep, complete=False)


def asMRRB(
    roots=None,
    r=(),
    c=(),
    complete=False,
):
    """
    Convenience Method to generate root_bunches with mirror real constraints
    from raw root lists
    """
    rb0 = RootBunch(r=r, c=c, constraint=root_constraints.mirror_real)

    if roots is not None:
        roots = np.asarray(roots)
        select_real = roots.imag == 0
        rr = roots[select_real]
        rc = roots[~select_real]
        if np.all(rc.imag < 0) and complete:
            return RootBunch(
                r=np.concatenate([rb0.r, rr]),
                c=np.concatenate([rb0.c, rc.conjugate()]),
                constraint=root_constraints.mirror_real,
            )
        elif np.all(rc.imag > 0) and complete:
            return RootBunch(
                r=np.concatenate([rb0.r, rr]),
                c=np.concatenate([rb0.c, rc]),
                constraint=root_constraints.mirror_real,
            )
        else:
            rb = ZPKTF.RBalgo.expect(roots, constraint=root_constraints.mirror_real)
            return rb * rb0
    else:
        return rb0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import sys
import time
import logging
import numpy as np

import contextlib
from wield import declarative
from wield.bunch import Bunch

from .. import fitters_ZPK
from .. import representations
from ..utilities.strings import padding_remove
from . import algorithms


class FitterUnacceptable(Exception):
    pass


class FitAid(object):
    def __init__(
        self,
        hints=None,
        hints_seen=None,
        active=True,
    ):
        self.fitter = None
        self.fitter_special = Bunch()
        self._fitter_current = None
        self._fitter_lowres_avg = None
        self._fitter_lowres_max = None
        self._fitter_lowres_med = None
        self._fitter_current_factorized = [None]
        self.N_update = 0

        self.mtime_start = time.time()

        if hints is None:
            hints = dict()

        if isinstance(hints, (list, tuple)):
            usehints = dict()
            for hdict in hints:
                usehints.update(hdict)
        else:
            usehints = dict(hints)

        self.hints = usehints
        self.hints_seen = hints_seen

        self.section = [0]
        self.section_stack = []

        # TODO make this a hint
        self.verbosity = 5
        # self.verbosity_info
        # self.verbosity_alert
        # self.verbosity_warn
        # self.verbosity_debug
        self.active = active

        # holds a heading for the logging, as well as sets tabbing
        self.log_header_stack = ()
        # indicates how far into the header has been printed yet.
        # for the live log
        self.log_header_printed = 0

        # log entries
        self._logs = []
        self.log_number = 0

        # stores all representative fitters seen and log metadata and checkpoint metadata
        self._fitters = []
        self._fitter_factors = []

        # stores all _checkpoints and metadata
        self._checkpoints = []
        self.checkpoint_current = None

        # investigations to view
        self.investigations = dict()
        return

    def __bool__(self):
        return self.active

    def __nonzero__(self):
        return self.active

    @property
    def residuals_type(self):
        return self.hint("residuals_type", "log")

    @property
    def residuals_type_alt(self):
        return self.hint("residuals_type_alt", "dualA")

    def _resolve_factorizations(self, fitter):
        """
        Get a fitter with representative residuals for comparison and storage.
        This applies all current factorizations.
        """
        if not isinstance(fitter, fitters_ZPK.MultiReprFilterBase):
            raise FitterUnacceptable()

        if self._fitter_factors:
            first_factor = self._fitter_factors[0]
            zeros = first_factor.zeros
            poles = first_factor.poles
            gain = first_factor.gain
            for factor in self._fitter_factors[1:]:
                zeros = zeros * factor.zeros
                poles = poles * factor.poles
                gain = gain * factor.gain
            xrep = representations.ZPKwData(
                ZPKrep=first_factor.ZPKrep,
                zeros=fitter.zeros * zeros,
                poles=fitter.poles * poles,
                gain=fitter.gain,
                delay_s=fitter.delay_s,
            )
            # TODO, check that the representation is preserved

            # print("residsB", fitter.residuals_average, len(fitter.poles) + len(fitter.poles_overlay))
            # from .. import plots
            # axB = plots.plot_fitter_flag(fitter=fitter, xscale='log')
            # axB.save("RESIDSB_{}.pdf".format(self.N_update))
            f2 = first_factor.regenerate(
                ZPKrep=xrep,
                residuals_log_im_scale=first_factor.residuals_log_im_scale,
                # needed to preserve the exact root locations
                #coding_map=fitters_ZPK.codings_s.coding_maps.SOS,
            )
            # print("residsB2", f2.residuals_average, len(f2.poles) + len(f2.poles_overlay))
            # axB = plots.plot_fitter_flag(fitter=f2, xscale='log')
            # axB.save("RESIDSB2_{}.pdf".format(self.N_update))
            return f2

        # TODO, should this assert if residuals_type is wrong?
        if fitter.residuals_type == self.residuals_type:
            if np.all(fitter.residuals_log_im_scale) != 1:
                f_new = fitter.copy(
                    residuals_type=self.residuals_type,
                )
                return f_new
        return None

    # declarative
    def fitter_orders(self, fitter=None):
        if fitter is None:
            fitter = self.fitter
        #if self._fitter_factors:
        #    first_factor = self._fitter_factors[0]
        #    total_z = len(first_factor.zeros) + len(first_factor.zeros_overlay)
        #    total_p = len(first_factor.poles) + len(first_factor.poles_overlay)
        #    for factor in self._fitter_factors[1:]:
        #        total_z += len(factor.zeros)
        #        total_p += len(factor.poles)
        #    factors_z = total_z
        #    factors_p = total_p
        #    if fitter != "factors":
        #        total_z += len(fitter.zeros)
        #        total_p += len(fitter.poles)
        #else:
        if fitter != "factors":
            factors_z = len(fitter.zeros_overlay)
            factors_p = len(fitter.poles_overlay)
            total_z = len(fitter.zeros) + len(fitter.zeros_overlay)
            total_p = len(fitter.poles) + len(fitter.poles_overlay)
        else:
            factors_z = len(self.fitter.zeros_overlay)
            factors_p = len(self.fitter.poles_overlay)
            total_z = len(self.fitter.zeros_overlay)
            total_p = len(self.fitter.poles_overlay)

        return Bunch(
            z=total_z,
            p=total_p,
            factors_z=factors_z,
            factors_p=factors_p,
            maxzp=max(total_z, total_p),
            factors_maxzp=max(factors_z, factors_p),
            factors_reldeg=factors_z - factors_p,
            total=total_z + total_p,
            reldeg=total_z - total_p,
        )

    # def hint_set(self, hname, hval, default = False):
    #    if default:
    #        self.hints.setdefault(hname, hval)
    #    else:
    #        self.hints[hname] = hval
    #    return

    def hint_has(self, hname):
        return hname in self.hints

    def hint_setdefault(self, hname, hval):
        self.hints.setdefault(hname, hval)
        return

    def hint_arg(self, func_arg, *args, **kwargs):
        if func_arg is None:
            return self.hint(*args, **kwargs)
        else:
            return func_arg

    def hint(self, *args, **kwargs):
        superarg = []
        for arg in args:
            if isinstance(arg, (list, tuple)):
                superarg.extend(arg)
            else:
                superarg.append(arg)

        # helper for when the known keys are being indexed
        if self.hints_seen is not None:
            keys_remapped = [key.format(**kwargs) for key in superarg]
            for idx, key in enumerate(keys_remapped):
                overrides, relateds = self.hints_seen.setdefault(key, (set(), set()))
                relateds.update(keys_remapped[:idx])
                overrides.update(keys_remapped[idx + 1 :])

        for key in superarg:
            key = key.format(**kwargs)
            ret = self.hints.get(key, declarative.NOARG)
            if ret is not declarative.NOARG:
                return ret
        return kwargs["default"]

    def fitter_check(
        self,
        fitter_new=None,
        hint_name=None,
        update=True,
        variant=None,
        validate=True,
    ):
        """ """

        if variant is None:
            raise RuntimeError(
                'Must specify variant for fitter_check, such as ["OrdUp", "OrdDn", "OrdC"]'
            )

        if self.fitter is None:
            if update:
                self.fitter = fitter_new
                self.fitter_update()
            else:
                return True

        if fitter_new is None:
            fitter_new = self.fitter

        # TODO, should check that it is using standard residuals

        resavgN = lambda: fitter_new.residuals_average
        resmaxN = lambda: fitter_new.residuals_max
        resmedN = lambda: fitter_new.residuals_med
        resavgC = lambda: self._fitter_current.residuals_average
        resavgC2 = lambda: self._fitter_current_factorized[-1].residuals_average if self._fitter_current_factorized[-1] is not None else -1
        resmaxC = lambda: self._fitter_current.residuals_max
        resmedC = lambda: self._fitter_current.residuals_med
        resavgB = lambda: self._fitter_lowres_avg.residuals_average
        resmaxB = lambda: self._fitter_lowres_max.residuals_max
        resmedB = lambda: self._fitter_lowres_med.residuals_med

        self.log_debug(9, "Residuals N:{0:.2f}, C:{1:.2f}, C2:{1:.2f}".format(resavgN(), resavgC(), resavgC2()))

        if hint_name is None:
            hint_name = "default"

        improved = True

        def check_hint(
            hints,
            calc,
        ):
            nonlocal improved
            hintval = self.hint(hints, default=None)
            if hintval is not None and calc > hintval:
                self.log_debug(10, "Failed {}, {} > {}".format(hints[0], calc, hintval))
                improved = False
            else:
                self.log_debug(12, "Succeed {}, {} < {}".format(hints[0], calc, hintval))

        check_hint(
            [
                "resavg_Ethresh{0}.{1}".format(variant, hint_name),
                "resavg_Ethresh{0}".format(variant),
            ],
            resavgN(),
        )

        check_hint(
            [
                "resmax_Ethresh{0}.{1}".format(variant, hint_name),
                "resmax_Ethresh{0}".format(variant),
            ],
            resmaxN(),
        )

        check_hint(
            [
                "resavg_Rthresh{0}.{1}".format(variant, hint_name),
                "resavg_Rthresh{0}".format(variant),
            ],
            (resavgN() / resavgC()),
        )

        check_hint(
            [
                "resmax_Rthresh{0}.{1}".format(variant, hint_name),
                "resmax_Rthresh{0}".format(variant),
            ],
            (resmaxN() / resmaxC()),
        )

        check_hint(
            [
                "resmed_Rthresh{0}.{1}".format(variant, hint_name),
                "resmed_Rthresh{0}".format(variant),
            ],
            (resmedN() / resmedC()),
        )

        check_hint(
            [
                "resavgB_Rthresh{0}.{1}".format(variant, hint_name),
                "resavgB_Rthresh{0}".format(variant),
            ],
            (resavgN() / resavgB()),
        )

        check_hint(
            [
                "resmaxB_Rthresh{0}.{1}".format(variant, hint_name),
                "resmaxB_Rthresh{0}".format(variant),
            ],
            (resmaxN() / resmaxB()),
        )

        check_hint(
            [
                "resmedB_Rthresh{0}.{1}".format(variant, hint_name),
                "resmedB_Rthresh{0}".format(variant),
            ],
            (resmedN() / resmedB()),
        )

        self.log_debug(
            10, "IMPROVED: ", improved, "residuals: ", resavgN(), resavgC()
        )

        val_func = self.hint("fitter_check_validate", default=None)
        if validate and val_func is not None and fitter_new is not None:
            val_func(self, fitter_new)

        if improved and update:
            self.fitter_update(
                fitter_new,
                representative=True,
            )
        return improved

    def fitter_checkup(self, variant=None, hint_name=None):
        ord_chg = self.fitter.order_total - self._fitter_current.order_total
        if variant is None:
            if ord_chg < 0:
                variant = "OrdDn"
            elif ord_chg > 0:
                variant = "OrdUp"
            else:
                variant = "OrdC"
        algorithms.sign_check_flip(self.fitter)
        improved = self.fitter_check(
            self.fitter,
            variant=variant,
            hint_name=hint_name,
        )
        if not improved:
            # Not sure I like this logic

            self.log_warn(3, "Fitter_checkup improvement fail")
            if self._fitter_current_factorized[-1] is not None:
                self.fitter = self._fitter_current_factorized[-1].copy()
        else:
            self.log_warn(3, "Fitter_checkup improvement succeed")
        return improved

    def fitter_checkpoint(self, variant=None, hint_name=None):
        """
        Just like fitter_checkup but doesn't revert the filter. It will make it representative if it is an improvement
        """
        ord_chg = self.fitter.order_total - self._fitter_current.order_total
        if variant is None:
            if ord_chg < 0:
                variant = "OrdDn"
            elif ord_chg > 0:
                variant = "OrdUp"
            else:
                variant = "OrdC"
        algorithms.sign_check_flip(self.fitter)
        improved = self.fitter_check(
            self.fitter,
            variant=variant,
            hint_name=hint_name,
        )

        if improved:
            self.log_warn(3, "Fitter_checkpoint improvement succeed, {}".format(hint_name))
            self.fitter_update(representative=True)
        return improved

    def fitter_update(
        self,
        fitter=None,
        representative=False,
        validate=True,
    ):
        self.N_update += 1
        if fitter is None:
            fitter = self.fitter
        else:
            self.fitter = fitter

        if self._fitter_x is not None:
            assert(len(self._fitter_x.poles_overlay) > 0)
            assert(len(self.fitter.poles_overlay) > 0)

        if self._fitter_factors and len(self._fitter_factors[-1].poles.fullplane) > 0:
            assert(len(fitter.poles_overlay) > 0)

        # print("fitter storeA", fitter.residuals_average)
        avg1 = fitter.residuals_average
        algorithms.sign_check_flip(fitter)
        avg2 = fitter.residuals_average
        assert avg2 <= avg1

        val_func = self.hint("fitter_update_validate", default=None)
        if validate and val_func is not None and fitter is not None:
            val_func(self, fitter)

        # this function will apply the de-factorizations, to create an
        # unfactored fit
        try:
            fitter_use = self._resolve_factorizations(fitter)
        except FitterUnacceptable:
            print("Unacceptable?!")
            return

        # print("ZPK", fitter_use.ZPKsf)
        if representative:
            if fitter_use is not None:
                fitter_use.optimize(aid=self)
                fitter_copy = [fitter_use]
            else:
                fitter_use = fitter
                fitter_copy = []

            def copy1():
                if not fitter_copy:
                    fitter_copy.append(fitter.copy())
                return fitter_copy[0]

            fitter_meta = Bunch()
            fitter_meta.fitter = copy1()
            fitter_meta.log_idx = self.log_number
            fitter_meta.checkpoint_idx = len(self._checkpoints) - 1
            fitter_meta.valid = True

            # These should possibly be run only if ensure_stable is in play
            # shouldn't be so necessary after previous debugging
            # assert(np.all(fitter.poles.fullplane.real < 0))
            # assert(np.all(fitter_meta.fitter.poles.fullplane.real < 0))
            self._fitters.append(fitter_meta)

            new_res = fitter_use.residuals_average
            self.log_progress(
                6,
                "New representative, Res: {:.2f}, from: {:.2f}".format(new_res, self.fitter.residuals_average),
            )
            new_res_max = fitter_use.residuals_max
            new_res_med = fitter_use.residuals_med

            self._fitter_current_factorized[-1] = fitter.copy()
            if self._fitter_current is not None:
                old_res = self._fitter_lowres_avg.residuals_average
                old_res_max = self._fitter_lowres_max.residuals_max
                old_res_med = self._fitter_lowres_med.residuals_med

                if new_res < old_res:
                    self._fitter_lowres_avg = copy1()

                if new_res_max < old_res_max:
                    self._fitter_lowres_max = copy1()

                if new_res_med < old_res_med:
                    self._fitter_lowres_med = copy1()

                self._fitter_current = copy1()
            else:
                self._fitter_current = copy1()
                self._fitter_lowres_avg = copy1()
                self._fitter_lowres_max = copy1()
                self._fitter_lowres_med = copy1()

        self.fitter_special = Bunch()
        return None

    def factorization_push(
        self,
        fitter_factor=None,
        data_mod=False,
    ):
        if fitter_factor is None:
            fitter_factor = self.fitter
        self._fitter_factors.append(fitter_factor)

        if data_mod:
            assert(False)
            xrep = representations.ZPKwData(
                ZPKrep=fitter_factor.ZPKrep,
                data=self.fitter.data / fitter_factor.xfer_fit,
                gain=1,
                poles=(),
                zeros=(),
                poles_overlay=(),
                zeros_overlay=(),
                delay_s=self.fitter.delay_s - fitter_factor.delay_s,
            )
            self.fitter = self.fitter.regenerate(ZPKrep=xrep)
        else:
            poles_overlay = fitter_factor.poles * fitter_factor.poles_overlay
            zeros_overlay = fitter_factor.zeros * fitter_factor.zeros_overlay
            xrep = representations.ZPKwData(
                ZPKrep=fitter_factor.ZPKrep,
                data=self.fitter.data,
                gain=fitter_factor.gain,
                poles=(),
                zeros=(),
                poles_overlay=poles_overlay,
                zeros_overlay=zeros_overlay,
                delay_s=self.fitter.delay_s - fitter_factor.delay_s,
            )
            self.fitter = self.fitter.regenerate(ZPKrep=xrep)
            self._fitter_current_factorized.append(None)
            if len(self.fitter.poles_overlay) > 0:
                self._fitter_x = self.fitter
            else:
                self._fitter_x = None
        return

    _fitter_x = None

    def factorization_pop(self):
        if self._fitter_factors and len(self._fitter_factors[-1].poles.fullplane) > 0:
            assert(len(self.fitter.poles_overlay) > 0)

        factorization = self._fitter_factors.pop()

        # print("resids", self.fitter.residuals_average, len(self.fitter.poles) + len(self.fitter.poles_overlay))
        # print("residsC", self._fitter_current.residuals_average)

        # from .. import plots
        # axB = plots.plot_fitter_flag(fitter=self.fitter, xscale='log')
        # axB.save("RESIDS_{}.pdf".format(self.N_update))
        self.fitter = self.fitter.regenerate(
            ZPKrep=factorization.ZPKrep,
            zeros=self.fitter.zeros * factorization.zeros,
            poles=self.fitter.poles * factorization.poles,
            gain=self.fitter.gain,
        )

        fcf = self._fitter_current_factorized.pop()
        if fcf is not None:
            # print("FCF Resids", fcf.residuals_average)
            self._fitter_current_factorized[-1] = fcf.regenerate(
                ZPKrep=factorization.ZPKrep,
                zeros=fcf.zeros * factorization.zeros,
                poles=fcf.poles * factorization.poles,
                gain=fcf.gain,
            )
            # print("FCF Resids2", self._fitter_current_factorized[-1].residuals_average)
            # print(fcf.poles, fcf.poles_overlay, fcf.gain)
            # fcf2 = self._fitter_current_factorized[-1]
            # print(fcf2.poles, fcf2.poles_overlay, fcf2.gain)
            # print("Zeros")
            # print(fcf.zeros, fcf.zeros_overlay, fcf.gain)
            # print(fcf2.zeros, fcf2.zeros_overlay, fcf2.gain)
            # print("done")

            # from wield.utilities.mpl import mplfigB
            # axB = mplfigB(Nrows=2)

            # h = fcf.xfer_fit / fcf2.xfer_fit
            # axB.ax0.loglog(fcf.F_Hz, abs(h), marker='.', ls='')
            # axB.ax1.semilogx(fcf.F_Hz, np.angle(h, deg=True), marker='.', ls='')
            # axB.save("fcf_tests_{}.pdf".format(len(self._fitter_current_factorized)))
            # print("FCF TEST {}".format(len(self._fitter_current_factorized)))

        # print("resids2?", self.fitter.residuals_average)
        # print("resids2C", self._fitter_current.residuals_average)

        # axB = plots.plot_fitter_flag(fitter=self.fitter, xscale='log')
        # axB.save("RESIDS2_{}.pdf".format(self.N_update))

        if self._fitter_factors and len(self._fitter_factors[-1].poles.fullplane) > 0:
            assert(len(self.fitter.poles_overlay) > 0)

        self._fitter_x = factorization
        if len(factorization.poles_overlay) > 0:
            self._fitter_x = factorization
        else:
            self._fitter_x = None
        if self._fitter_x is not None:
            assert(len(self._fitter_x.poles_overlay) > 0)
            assert(len(self.fitter.poles_overlay) > 0)
        # TODO, check that the representation is preserved

    @contextlib.contextmanager
    def factorization(self, *args, **kwargs):
        self.factorization_push(*args, **kwargs)
        yield
        self.factorization_pop()

    def checkpoint(self, checkpoint):
        metadata = Bunch()
        metadata.name = checkpoint
        metadata.log_idx = self.log_number
        metadata.fitter_idx = self(self._fitters) - 1
        metadata.fitter_current = self._fitter_current
        metadata.fitter_lowres_avg = self._fitter_lowres_avg
        metadata.fitter_lowres_max = self._fitter_lowres_max
        metadata.fitter_lowres_med = self._fitter_lowres_med
        metadata.fitter = self.fitter.copy()
        metadata.fitter_special = Bunch()
        for name, fitter in self.fitter_special.items():
            metadata.fitter_special[name] = fitter.copy()
        metadata.fitter_factors = list(self._fitter_factors)
        self._checkpoints.append(metadata)

    def invalidate_fitters(self):
        """
        If an algorithm a
        TODO
        """
        for fmeta in self._fitters:
            fmeta.valid = False
        return

    def log(self, *args, **kwargs):
        """
        First argument is the level, should include a log group, which must be one of
        ['info', 'debug', 'warn', 'alert', 'rationale', 'progress']
        """
        level = args[0]
        if isinstance(level, int):
            level = args[0]
            args = args[1:]
            group = kwargs.setdefault("group", "info")
        else:
            level = -1
            group = kwargs.setdefault("group", "debug")
            # TODO print line and file upon hint request
            # args = args

        header = self.log_header_stack
        kwargs["header"] = header
        kwargs["time"] = time.time()
        kwargs["time_start"] = self.mtime_start

        if self.hint("log_off", default=False):
            return

        kwargs["args"] = args
        # TODO, merge if consecutive with the same parameters
        self._logs.append(kwargs)
        self.log_number += 1

        # FOR LIVE PRINTING
        if group == "info":
            log_mod_level = logging.INFO
            group_character = "I"
            level_limit = self.hint(
                [
                    "log_level_info",
                    "log_level",
                ],
                default=8,
            )
        elif group == "debug":
            log_mod_level = logging.DEBUG
            group_character = "D"
            level_limit = self.hint(
                [
                    "log_level_debug",
                    "log_level",
                ],
                default=8,
            )
        elif group == "warn":
            log_mod_level = logging.WARNING
            group_character = "W"
            level_limit = self.hint(
                [
                    "log_level_warn",
                    "log_level",
                ],
                default=8,
            )
        elif group == "alert":
            log_mod_level = logging.WARNING
            group_character = "A"
            level_limit = self.hint(
                [
                    "log_level_alert",
                    "log_level",
                ],
                default=8,
            )
        elif group == "rationale":
            log_mod_level = logging.INFO
            group_character = "R"
            level_limit = self.hint(
                [
                    "log_level_rationale",
                    "log_level",
                ],
                default=8,
            )
        elif group == "progress":
            log_mod_level = logging.INFO
            group_character = "P"
            level_limit = self.hint(
                [
                    "log_level_progress",
                    "log_level",
                ],
                default=8,
            )
        else:
            raise RuntimeError("Unrecognized log grouping")

        if self.hint("log_print", default=True) and level <= level_limit:
            hint_log_stdout = self.hint("log_stdout", default=True)
            if hint_log_stdout not in [None, True, False]:
                lfile = hint_log_stdout
            else:
                lfile = sys.stdout

            header = self.log_header_stack
            header_len = len(header)

            prefix = "{}{} {: >6.2f} {}".format(
                level if level >= 0 else "-",
                group_character,
                kwargs["time"] - kwargs["time_start"],
                "  " * header_len,
            )

            # TODO, make these take a header argument
            if not self.hint("logging_use", default=False):

                def pfunc(*args, **kwargs):
                    print(*args, **kwargs)

            else:

                def pfunc(*args, **kwargs):
                    kwargs.pop("file", None)
                    logging.log(log_mod_level + 9 - level, *args, **kwargs)

            logfile = self.hint("tee_logfile", default=None)
            if logfile is not None:
                def pfuncwrap(pfunc):
                    def pfunc2(*args, file=None, **kwargs):
                        """
                        wrapper to print as-is as well as to tee into the logfile
                        the file argument is eaten and relayed into the existing pfunc
                        """
                        pfunc(*args, file=file, **kwargs)
                        with open(logfile, 'a') as F:
                            print(*args, file=F, **kwargs)
                    return pfunc2
                pfunc = pfuncwrap(pfunc)

            if header_len > self.log_header_printed:
                pfunc(
                    "{}:{}:".format("-" * (len(prefix)), ":".join(header)), file=lfile
                )
                self.log_header_printed = header_len
                # tag that the header has been printed

            hint_log_stderr = self.hint("log_stderr", default=True)
            if hint_log_stderr and group == "warn":
                if hint_log_stderr not in [None, True, False]:
                    lfile = hint_log_stderr
                else:
                    lfile = sys.stderr
            else:
                lfile = sys.stdout

            arg_lines = [[]]
            for arg in args:
                if isinstance(arg, str):
                    if "\n" in arg:
                        arg = padding_remove(arg)
                    arg_spl = arg.split("\n")
                    arg_lines[-1].append(arg_spl[0])
                    for subline in arg_spl[1:]:
                        arg_lines.append([subline])
                else:
                    arg_lines[-1].append(arg)

            # TODO, have pfunc do this splitting
            pfunc(prefix, *arg_lines[0], file=lfile)
            for argsl in arg_lines[1:]:
                pfunc(" " * len(prefix), *argsl, file=lfile)
        return

    def log_debug(self, *args, **kwargs):
        kwargs["group"] = "debug"
        self.log(*args, **kwargs)

    def log_warn(self, *args, **kwargs):
        kwargs["group"] = "warn"
        self.log(*args, **kwargs)

    def log_alert(self, *args, **kwargs):
        kwargs["group"] = "alert"
        self.log(*args, **kwargs)

    def log_info(self, *args, **kwargs):
        kwargs["group"] = "info"
        self.log(*args, **kwargs)

    def log_rationale(self, *args, **kwargs):
        kwargs["group"] = "rationale"
        self.log(*args, **kwargs)

    def log_progress(self, *args, **kwargs):
        kwargs["group"] = "progress"
        self.log(*args, **kwargs)

    @contextlib.contextmanager
    def log_heading(self, header):
        save_stack = self.log_header_stack
        self.log_header_stack = save_stack + (header,)
        # TODO, auto print header on command?
        yield
        self.log_header_stack = save_stack
        if self.log_header_printed > len(save_stack):
            self.log_header_printed = len(save_stack)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from wield.bunch import Bunch

from .. import fitters_ZPK
from .. import annotate
from .. import plots
from .. import TFmath
from ..utilities import args

from . import order_reduce
from . import disc_sequence
from . import disc_sequence_mag
from . import fit_aid
from . import algorithms


def data2filter_setup(
    argB=args.UNSPEC,
    data=args.UNSPEC,
    F_Hz=args.UNSPEC,
    SNR=args.UNSPEC,
    SNR_initial=args.UNSPEC,
    SNR_cutoff=args.UNSPEC,
    F_nyquist_Hz=args.UNSPEC,
    ZPK=args.UNSPEC,
    ZPK_overlay=args.UNSPEC,
    order_initial=args.UNSPEC,
    verbosity_limit=args.UNSPEC,
    delay_s=args.UNSPEC,
    aid=args.UNSPEC,
    dropnans=args.UNSPEC,
    select=args.UNSPEC,
    complete=args.UNSPEC,
    hints=None,
    doc_db=None,
):
    data = args.argscan(locals(), argB, args.REQ, arg="data")
    F_Hz = args.argscan(locals(), argB, args.REQ, arg="F_Hz")
    SNR = args.argscan(locals(), argB, 1, arg="SNR")
    SNR_initial = args.argscan(locals(), argB, None, arg="SNR_initial")
    SNR_cutoff = args.argscan(locals(), argB, 100, arg="SNR_cutoff")
    F_nyquist_Hz = args.argscan(locals(), argB, 0, arg="F_nyquist_Hz")
    ZPK = args.argscan(locals(), argB, None, arg="ZPK")
    ZPK_overlay = args.argscan(locals(), argB, None, arg="ZPK_overlay")
    order_initial = args.argscan(locals(), argB, None, arg="order_initial")
    verbosity_limit = args.argscan(locals(), argB, 5, arg="verbosity_limit")
    delay_s = args.argscan(locals(), argB, 0, arg="delay_s")
    aid = args.argscan(locals(), argB, None, arg="aid")
    dropnans = args.argscan(locals(), argB, True, arg="dropnans")
    select = args.argscan(locals(), argB, None, arg="select")
    complete = args.argscan(locals(), argB, True, arg="complete")

    if delay_s == "fit":
        delay_s = None

    if F_nyquist_Hz == 0:
        F_nyquist_Hz = None

    if aid is None:
        # TODO make annotate_setup
        doc_db = annotate.create(
            doc_db, name="fit_sequence version 1", verbosity_limit=verbosity_limit
        )
        aid = fit_aid.FitAid(
            doc_db=doc_db,
            hints=hints,
        )
    else:
        doc_db = aid.doc_db

    SNR = np.minimum(SNR, SNR_cutoff)

    if SNR_initial is None:
        if SNR is None:
            SNR_initial = 1
        else:
            SNR_initial = SNR
    else:
        SNR_initial = np.minimum(SNR_initial, SNR_cutoff)

    if select is not None:
        (F_Hz, data, SNR, SNR_initial) = np.broadcast_arrays(
            F_Hz,
            data,
            SNR,
            SNR_initial,
        )
        F_Hz = F_Hz[select]
        data = data[select]
        SNR = SNR[select]
        SNR_initial = SNR_initial[select]

    if dropnans:
        nan_select = np.isfinite(data)
        if not np.all(nan_select):
            (F_Hz, data, SNR, SNR_initial,) = np.broadcast_arrays(
                F_Hz,
                data,
                SNR,
                SNR_initial,
            )
            F_Hz = F_Hz[nan_select]
            data = data[nan_select]
            SNR = SNR[nan_select]
            SNR_initial = SNR_initial[nan_select]
    return Bunch(locals())


def data2filter(argB=args.UNSPEC, mag_only=args.UNSPEC, alt_res=args.UNSPEC, **kwargs):
    mag_only = args.argscan(locals(), argB, False, arg="mag_only")
    alt_res = args.argscan(locals(), argB, False, arg="alt_res")
    setup = data2filter_setup(argB, **kwargs)

    aid = setup.aid

    aid.hint_set("resavg_RthreshOrdDn", 1.10, default=True)
    aid.hint_set("resavg_RthreshOrdUp", 1.00, default=True)
    aid.hint_set("resavg_RthreshOrdC", 1.00, default=True)
    aid.hint_set("overlay_native", False, default=True)

    if mag_only:
        aid.hint_set("residuals_type", "log", default=True)
        aid.hint_set("residuals_type_alt", "log", default=True)
    elif not alt_res:
        aid.hint_set("residuals_type", "log", default=True)
        aid.hint_set("residuals_type_alt", "dualA", default=True)
    else:
        aid.hint_set("residuals_type", "dualA", default=True)
        aid.hint_set("residuals_type_alt", "log", default=True)

    aid.annotate(
        name=None,
        method="v1.fit_sequence",
        about=(
            """
            Version 1 smart fitter in wield.iirrational library.
            Uses SVD method with high order over-fitting,
            then switches to nonlinear fits with heuristics
            to remove poles and zeros down to a reasonable system order.
            """
        ),
        verbosity=3,
    )
    MRFkwargs = dict(
        residuals_type=aid.hint("residuals_type", default="log"),
    )
    if mag_only:
        MRFkwargs["residuals_log_im_scale"] = 0
    else:
        pass

    overlay_native = aid.hint("overlay_native", default=False)

    if setup.ZPK is None:
        if not mag_only:
            ratdisc = disc_sequence.rational_disc_fit
        else:
            ratdisc = disc_sequence_mag.rational_disc_fit_mag

        if not overlay_native and setup.ZPK_overlay is not None:
            ZPK_overlay_hold = setup.ZPK_overlay
            ZPK_overlay = ((), (), 1)
            data_hold = setup.data
            data = data_hold / TFmath.TF_ZPK(
                F_Hz=setup.F_Hz,
                ZPK=ZPK_overlay_hold,
                F_nyquist_Hz=setup.F_nyquist_Hz,
            )
        else:
            data = setup.data
            ZPK_overlay = setup.ZPK_overlay

        with aid.annotate_into("initial"):
            rat_fitter = ratdisc(
                F_Hz=setup.F_Hz,
                data=data,
                SNR=setup.SNR_initial,
                nyquist_final_Hz=setup.F_nyquist_Hz,
                order=setup.order_initial,
                ZPK_overlay=ZPK_overlay,
                aid=aid,
                complete=setup.complete,
            )
        # secret output to log the rational fit output

        if not overlay_native and ZPK_overlay is not None:
            Zov, Pov, Kov = ZPK_overlay_hold
            rat_fitter.zeros_overlay = Zov
            rat_fitter.poles_overlay = Pov
            rat_fitter.gain = rat_fitter.gain * Kov
            rat_fitter.data = data_hold

        fitter = fitters_ZPK.ZPKrep2MRF(
            rat_fitter.ZPKrep,
            delay_hide=True,
            SNR=setup.SNR,
            coding_map=fitters_ZPK.coding_maps.nlFBW_safe,
            **MRFkwargs
        )

        validate = aid.hint("rational_fitter_validate", default=None)
        if validate is not None:
            validate(rat_fitter, fitter)

        # TODO, add documentation
        aid.fitter_update(fitter)
    else:
        # TODO, just use ZPKrep constructor
        aid.fitter_update(
            fitters_ZPK.ZPKdata2MRF(
                F_Hz=setup.F_Hz,
                data=setup.data,
                SNR=setup.SNR,
                ZPK=setup.ZPK,
                ZPK_overlay=setup.ZPK_overlay,
                F_nyquist_Hz=setup.F_nyquist_Hz,
                delay_hide=True,
                coding_map=fitters_ZPK.coding_maps.nlFBW_safe,
                **MRFkwargs
            )
        )

    aid.log(
        "Initial Order: (Z= {0}, P= {1}, Z-P= {2})".format(
            len(aid.fitter.zeros),
            len(aid.fitter.poles),
            len(aid.fitter.zeros) - len(aid.fitter.poles),
        )
    )

    duals = order_reduce.match_pairs(aid=aid)
    # do a pre-reduce for speed with very conservative cutoff
    ORb = order_reduce.order_reduce(
        aid=aid,
        duals=duals,
        Q_rank_cutoff=0.6,
    )
    aid.fitter_update(ORb.fitter_out)

    opt_r = aid.fitter.optimize(
        residuals_type=aid.hint("residuals_type_alt", default="log"),
        aid=aid,
    )
    aid.fitter_update()

    opt_r = aid.fitter.optimize(aid=aid)
    aid.fitter_update()

    aid.annotate(
        name="nonlinear pre-reduce and optimize",
        fitter=aid.fitter,
        plotter=plots.plot_fitter_flag,
        method="MultiReprFilterZ.optimize",
        about=(
            """
            initial conservative order reduction
            (for speed), followed by a nonlinear optimization.
            """
        ),
        verbosity=3,
    )

    ##TODO broken
    # if aid.hint('order_increase', default = True):
    #    improved = algorithms.insert_triplets(aid = aid)
    #    if improved:
    #        improved = algorithms.insert_triplets(aid = aid)

    opt_r = aid.fitter.optimize(aid=aid)
    aid.fitter_update()

    algorithms.set_min_BW(aid=aid)

    opt_r = aid.fitter.optimize(aid=aid)
    aid.fitter_update()

    aid.annotate(
        name="optimize nonlinear after bandwidth limiting",
        fitter=aid.fitter,
        plotter=plots.plot_fitter_flag,
        method="MultiReprFilterZ.optimize",
        about=(
            """
            Root bandwidths limited in the nonlinear representation
            to half of the local average the frequency spacing.
            Nonlinear optimization then applied.
            """
        ),
        verbosity=7,
    )

    with aid.annotate_into("order reduce 1"):
        order_reduce.order_reduce_aggressive(aid=aid)

    if aid.fitter.F_nyquist_Hz is not None:
        with aid.annotate_into("reduce_negroots"):
            order_reduce.order_reduce_negroots(aid=aid)
    aid.fitter_update()

    with aid.annotate_into("remove_weakroots"):
        improved = True
        while improved:
            improved = order_reduce.order_reduce_weakroots(
                aid=aid,
                delay_s=setup.delay_s,
            )
            aid.fitter_update()

    if aid.fitter.F_nyquist_Hz is not None:
        with aid.annotate_into("reduce_negroots"):
            order_reduce.order_reduce_negroots(aid=aid)

    opt_r = aid.fitter.optimize(
        aid=aid,
        residuals_type=aid.hint("residuals_type_alt", default="log"),
    )
    aid.fitter_update()

    opt_r = aid.fitter.optimize(
        aid=aid,
    )
    aid.fitter_update()

    if aid.fitter.F_nyquist_Hz is not None:
        with aid.annotate_into("reduce_negroots"):
            order_reduce.order_reduce_negroots(aid=aid)

    if setup.delay_s is None:
        aid.fitter.optimize_delay(aid=aid)
        aid.fitter_update()

    if aid.hint("flip_mindelay", default=True):
        with aid.annotate_into("flip_mindelay"):
            algorithms.flip_mindelay_opt(aid=aid)

    aid.fitter.optimize(aid=aid)
    aid.fitter_update()

    if setup.delay_s is None:
        aid.fitter.optimize_delay(aid=aid)
        aid.fitter_update()

    # final check that the sign is right..
    algorithms.sign_check_flip(aid.fitter)
    aid.fitter.optimize(aid=aid)
    aid.fitter_update()

    if setup.delay_s is None:
        aid.log("DELAY: ", aid.fitter.gain_delay_coding.p_delay)

    aid.log("FINAL RESIDUALS", aid.fitter.residuals_average)

    validate = aid.hint("fitter_final_validate", default=None)
    if validate is not None:
        validate(rat_fitter)
    # TODO, check sign and flip

    aid.annotate(
        name="final",
        fitter=aid.fitter,
        plotter=plots.plot_fitter_flag,
        method="MultiReprFilter.optimize",
        about=(
            """
            Optimizes using nonlinear parameterizations,
            """
        ),
        verbosity=0,
    )

    return aid



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

from wield.iirrational.testing import IIRrational_data
from wield.iirrational import testing

from wield.iirrational.fitters_ZPK import ZPKrep2MRF
from wield.iirrational.fitters_ZPK import codings_s

from utilities import check_residuals


@pytest.mark.parametrize("coding_map", sorted(codings_s.coding_maps.keys()))
def test_delay_fitter(coding_map):
    """
    tests when the fitter is bad as it is trying to fit on the wrong side
    """
    data = IIRrational_data("delay")
    fitterMRF = ZPKrep2MRF(
        data.rep_s.ZPKrep,
        coding_map=codings_s.coding_maps[coding_map],
    )
    fitterMRF.delay_s = 0.000000
    fitterMRF.codings_revision += fitterMRF.codings_revision + 1
    testing.assert_almost_equal(data.rep_s.xfer_fit / fitterMRF.xfer_fit, 1, 4)
    print("DELAY MRF: ", fitterMRF.delay_s)
    print("DELAY DATA: ", data.rep_s.delay_s)
    print("DELAY MRF MIN: ", fitterMRF.delay_s_min)
    print("DELAY MRF MAX: ", fitterMRF.delay_s_max)

    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    opt = fitterMRF.optimize()
    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    assert fitterMRF.residuals_average < 2
    # axB = plot_fitter_flag(fitterMRF, with_error=False, plot_zp = True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from matplotlib import gridspec
import matplotlib.pyplot as plt

from .. import fitters_ZPK

from ..utilities.mpl import (
    asavefig,
    generate_ax,
)


def plot_fitter_flag(
    self,
    fitter,
    title=None,
    plot_zp=True,
    plot_to_nyquist=True,
    with_error=True,
    with_data_error=True,
    N_points=300,
    gs_base=gridspec.GridSpec(1, 1)[0],
    fig=None,
    ax_bunch=None,
    xscale=None,
    plot_data_scale=10,
    plot_data_scale_low=2,
):
    if not isinstance(fitter, fitters_ZPK.MultiReprFilterBase):
        fitter = fitters_ZPK.ZPKrep2MRF(
            fitter.ZPKrep,
            coding_map=fitters_ZPK.coding_maps.RI,
        )
    axB = generate_ax(ax_bunch)

    if fitter.F_nyquist_Hz is None or fitter.F_nyquist_Hz > 1.2 * fitter.F_max_Hz:
        if fig is None:
            fig = plt.figure()
            fig.set_size_inches(12, 9)
            axB.fig = fig
        gs_DC = gridspec.GridSpecFromSubplotSpec(
            2,
            2,
            subplot_spec=gs_base,
            wspace=0.2,
            hspace=0.2,
            # height_ratios = height_ratio_list,
            width_ratios=[1, 0.3],
        )

        gs_data = gs_DC[0, 0]
        gs_nyquist = gs_DC[1, 0]
        gs_zp = gs_DC[:, 1]
        zp_horz = False
    else:
        if fig is None:
            fig = plt.figure()
            fig.set_size_inches(10, 9)
            axB.fig = fig
        gs_DC = gridspec.GridSpecFromSubplotSpec(
            2,
            1,
            subplot_spec=gs_base,
            wspace=0.1,
            hspace=0.3,
            height_ratios=[1, 0.7],
        )

        gs_data = gs_DC[0, 0]
        gs_nyquist = False
        gs_zp = gs_DC[1, 0]
        zp_horz = True

    if title is not None:
        axB.fig.suptitle(title)

    if gs_data:
        axB_top = self.plot_fit(
            fitter=fitter,
            ax_bunch=None,
            gs_base=gs_data,
            with_error=with_error,
            with_data_error=with_data_error,
            plot_to_nyquist=False,
            plot_now=False,
            xscale=xscale,
            fig=fig,
        )
        axB.top = axB_top
        axB_top.plot_data(fitter)
        axB_top.plot_fit(fitter)
        axB_top.plot_zp(fitter)
        axB.finalizers.append(axB.top.finalize)

    if gs_nyquist:
        axB_bottom = self.plot_fit(
            fitter=fitter,
            ax_bunch=None,
            gs_base=gs_nyquist,
            with_error=False,
            with_data_error=False,
            plot_past_data=True,
            plot_to_nyquist=True,
            plot_now=False,
            fig=fig,
            xscale="log",
            plot_data_scale=plot_data_scale,
            plot_data_scale_low=plot_data_scale_low,
        )
        axB.bottom = axB_bottom
        axB_bottom.plot_data(fitter)
        axB_bottom.plot_fit(fitter)
        axB_bottom.plot_zp(fitter)
        left, right = axB.bottom.ax_bottom.get_xlim()
        axB.finalizers.append(axB.bottom.finalize)
        if fitter.F_nyquist_Hz is not None:
            axB.bottom.ax_bottom.set_xlim(left, fitter.F_nyquist_Hz)

    if gs_zp:
        if fitter.F_nyquist_Hz is None or fitter.F_nyquist_Hz / max(fitter.F_Hz) > 5:
            axB_right = self.plot_ZP_S(
                fitter=fitter,
                ax_bunch=None,
                gs_base=gs_zp,
                horizontal=zp_horz,
                fig=fig,
            )
        else:
            axB_right = self.plot_ZP(
                fitter=fitter,
                ax_bunch=None,
                gs_base=gs_zp,
                horizontal=zp_horz,
                fig=fig,
            )
        axB.right = axB_right

    def plot_fit(*args, **kwargs):
        db = axB.top.plot_fit(*args, **kwargs)
        axB.bottom.plot_fit(*args, db_ref=db, **kwargs)

    axB.plot_fit = plot_fit

    def save(rootname, **kwargs):
        axB.finalize()
        axB << asavefig(rootname, **kwargs)

    axB.save = save
    axB.finalize()
    return axB


def plot_fitter_flag_compare(
    self,
    fitter,
    fitter_ref,
    N_points=300,
    gs_base=gridspec.GridSpec(1, 1)[0],
    fig=None,
    ax_bunch=None,
):
    if not isinstance(fitter, fitters_ZPK.MultiReprFilterBase):
        fitter = fitters_ZPK.RDF2MRF(fitter)
    if not isinstance(fitter_ref, fitters_ZPK.MultiReprFilterBase):
        fitter_ref = fitters_ZPK.RDF2MRF(fitter_ref)
    axB = generate_ax(ax_bunch)

    if fig is None:
        fig = plt.figure()
        fig.set_size_inches(14, 9)
        axB.fig = fig
    gs_DC = gridspec.GridSpecFromSubplotSpec(
        3,
        3,
        subplot_spec=gs_base,
        wspace=0.3,
        hspace=0.2,
        height_ratios=[1, 1, 0.5],
        width_ratios=[1, 0.3, 0.3],
    )

    gs_data = gs_DC[0, 0]
    gs_nyquist = gs_DC[1, 0]
    gs_zp = gs_DC[:2, 1]
    gs_zp_ref = gs_DC[:2, 2]
    zp_horz = False
    gs_rel = gs_DC[2, :]

    if gs_data:
        axB_top = self.plot_fit(
            fitter=fitter,
            gs_base=gs_data,
            with_error=True,
            with_data_error=True,
            plot_to_nyquist=False,
            fig=fig,
            ax_bunch=None,
        )
        axB.top = axB_top
        axB_top = self.plot_fit(
            fitter=fitter_ref,
            gs_base=gs_data,
            with_error=False,
            with_data_error=False,
            plot_data=False,
            plot_to_nyquist=False,
            fig=fig,
            ax_bunch=axB_top,
            plot_zp=False,
            label_fit="original ({nP}P, {nZ}Z, Rsq={Rsq:.2e})",
        )
        axB_top.mag.legend(ncol=1, fontsize=8, loc="best").set_visible(False)

    if gs_nyquist:
        axB_bottom = self.plot_fit(
            fitter=fitter,
            ax_bunch=None,
            gs_base=gs_nyquist,
            with_error=False,
            with_data_error=False,
            plot_to_nyquist=True,
            fig=fig,
            xscale="log",
        )
        axB_bottom = self.plot_fit(
            fitter=fitter_ref,
            gs_base=gs_nyquist,
            with_error=False,
            with_data_error=False,
            plot_data=False,
            plot_to_nyquist=True,
            fig=fig,
            xscale="log",
            ax_bunch=axB_bottom,
            plot_zp=False,
            label_fit="original ({nP}P, {nZ}Z, Rsq={Rsq:.2e})",
        )
        axB.bottom = axB_bottom

        def finalize():
            axB_bottom.mag.legend(ncol=1, fontsize=8, loc="best")
            left, right = axB.bottom.ax_bottom.get_xlim()
            axB.bottom.ax_bottom.set_xlim(left, fitter.F_nyquist_Hz)

        axB_bottom.finalizers.append(finalize)

    if gs_zp:
        if fitter.F_nyquist_Hz / max(fitter.F_Hz) > 5:
            axB_right = self.plot_ZP_S(
                fitter=fitter,
                ax_bunch=None,
                gs_base=gs_zp,
                horizontal=zp_horz,
                fig=fig,
            )
        else:
            axB_right = self.plot_ZP(
                fitter=fitter,
                ax_bunch=None,
                gs_base=gs_zp,
                horizontal=zp_horz,
                fig=fig,
            )
        axB.right = axB_right
        axB.right.ax_poles.set_title("fit roots (s)")

    if gs_zp_ref:
        if fitter.F_nyquist_Hz / max(fitter.F_Hz) > 5:
            axB_right = self.plot_ZP_S(
                fitter=fitter_ref,
                ax_bunch=None,
                gs_base=gs_zp_ref,
                horizontal=zp_horz,
                fig=fig,
            )
        else:
            axB_right = self.plot_ZP(
                fitter=fitter_ref,
                ax_bunch=None,
                gs_base=gs_zp_ref,
                horizontal=zp_horz,
                fig=fig,
            )
        axB.rright = axB_right
        axB.rright.ax_poles.set_title("reference roots (s)")

    if gs_rel:
        self.plot_rel_comparison(
            fitter=fitter,
            fitter_ref=fitter_ref,
            gs_base=gs_rel,
            fig=fig,
        )

    def save(rootname, **kwargs):
        axB << asavefig(rootname, **kwargs)

    axB.save = save

    axB.finalize()
    return axB


def plot_fitter_flag_residuals(
    self,
    fitter,
    title=None,
    plot_zp=True,
    plot_to_nyquist=True,
    with_error=True,
    with_data_error=True,
    N_points=300,
    gs_base=gridspec.GridSpec(1, 1)[0],
    fig=None,
    ax_bunch=None,
    xscale=None,
):
    if not isinstance(fitter, fitters_ZPK.MultiReprFilterBase):
        fitter = fitters_ZPK.ZPKrep2MRF(
            fitter.ZPKrep,
            coding_map=fitters_ZPK.coding_maps.RI,
        )
    axB = generate_ax(ax_bunch)

    if fig is None:
        fig = plt.figure()
        fig.set_size_inches(12, 9)
        axB.fig = fig

    gs_DC = gridspec.GridSpecFromSubplotSpec(
        2,
        1,
        subplot_spec=gs_base,
        wspace=0.2,
        hspace=0.2,
    )

    gs_data = gs_DC[0, 0]
    gs_res = gs_DC[1, 0]

    if title is not None:
        axB.fig.suptitle(title)

    if gs_data:
        axB_top = self.plot_fit(
            fitter=fitter,
            ax_bunch=None,
            gs_base=gs_data,
            with_error=with_error,
            with_data_error=with_data_error,
            plot_past_data=True,
            plot_to_nyquist=True,
            plot_now=False,
            fig=fig,
            xscale=xscale,
        )
        axB_top.plot_data(fitter)
        db_fit = axB_top.plot_fit(fitter)
        axB_top.plot_zp(fitter)
        axB.top = axB_top
        axB.finalizers.append(axB.top.finalize)

    if gs_res:
        #axB_bottom = self.plot_residuals(
        #    fitter=fitter,
        #    ax_bunch=None,
        #    gs_base=gs_res,
        #    plot_now=False,
        #    fig=fig,
        #    xscale=xscale,
        #)
        axB_bottom = self.plot_fit(
            fitter=fitter,
            ax_bunch=None,
            gs_base=gs_res,
            with_error=with_error,
            with_data_error=with_data_error,
            plot_past_data=False,
            plot_to_nyquist=True,
            plot_now=False,
            residuals=True,
            fig=fig,
            xscale=xscale,
        )
        axB.bottom = axB_bottom
        axB_bottom.mag.axhline(1.1, ls='--', color='black', lw=1)
        axB_bottom.mag.axhline(1.01, ls='--', color='black', lw=1)
        axB_bottom.mag.axhline(1/1.1, ls='--', color='black', lw=1)
        axB_bottom.mag.axhline(1/1.01, ls='--', color='black', lw=1)
        axB_bottom.plot_data(fitter)
        axB_bottom.plot_fit(fitter, db_ref=db_fit)
        left, right = axB.bottom.ax_bottom.get_xlim()
        if fitter.F_nyquist_Hz is not None:
            axB.bottom.ax_bottom.set_xlim(left, fitter.F_nyquist_Hz)
        axB.finalizers.append(axB.bottom.finalize)

    def plot_fit(*args, **kwargs):
        db = axB.top.plot_fit(*args, **kwargs)
        axB.bottom.plot_fit(*args, db_ref=db, **kwargs)

    axB.plot_fit = plot_fit

    def save(rootname, **kwargs):
        axB.finalize()
        axB << asavefig(rootname, **kwargs)

    axB.save = save
    axB.finalize()
    return axB



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
from wield.bunch import Bunch

from ... import TFmath
from ... import representations

from ...utilities import ensure_aid
from ...fitters_ZPK.ZPKrep2MRF import ZP2codings


def resrank_program(
    fitter,
    rank_zp_idx_list,
    name,
    program,
    variant=None,
    marginalize_delay=True,
    **kwargs
):

    Pc = fitter.poles.c
    Zc = fitter.zeros.c
    Pr = fitter.poles.r
    Zr = fitter.zeros.r

    Pc_mod = []
    Pr_mod = []
    Zc_mod = []
    Zr_mod = []

    program_copy = list(program)
    while program_copy:
        arg = program_copy.pop()
        which = program_copy.pop()
        if which == "PcDelIdx":
            Pc_mod.append(Pc[arg])
        elif which == "ZcDelIdx":
            Zc_mod.append(Zc[arg])
        elif which == "PrDelIdx":
            Pr_mod.append(Pr[arg])
        elif which == "ZrDelIdx":
            Zr_mod.append(Zr[arg])
        # swap on these since they are added
        elif which == "PrAdd":
            Zr_mod.append(arg)
        elif which == "ZrAdd":
            Pr_mod.append(arg)
        elif which == "PcAdd":
            Zc_mod.append(arg)
        elif which == "ZcAdd":
            Pc_mod.append(arg)

    PrB = representations.asMRRB(c=Pc_mod, r=Pr_mod)
    ZrB = representations.asMRRB(c=Zc_mod, r=Zr_mod)
    h, lnG = PrB.val_lnG(fitter.F_Hz)
    h, lnG = ZrB.val_lnG(fitter.F_Hz, h=1 / h, lnG=-lnG)

    R = fitter.xfer_fit / (fitter.data * h)

    # this is the exact solution for the gain adjustment required, using the DualB
    # residuals
    G = (
        np.sum(fitter.W ** 2 / TFmath.abs_sq(R))
        / np.sum(fitter.W ** 2 * TFmath.abs_sq(R))
    ) ** 0.25
    R = G * R

    rank = np.sum(TFmath.abs_sq(fitter.residuals_NLmap(R, W=fitter.W)))
    if marginalize_delay:
        # linear detrends the phasing term to remove delay effects
        f = fitter.F_Hz
        # the factor of 2 is not with the exact solution, but seems to work better
        R.imag -= (
            f * np.sum(f * R.imag * fitter.W ** 2) / np.sum(f ** 2 * fitter.W ** 2) / 2
        )
        rank_moddelay = np.sum(TFmath.abs_sq(fitter.residuals_NLmap(R, W=fitter.W)))
        rank = min(rank, rank_moddelay)
    pbunch = Bunch(**kwargs)
    pbunch.rank = rank
    pbunch.name = name
    pbunch.variant = variant
    pbunch.program = program
    pbunch.flip = True
    rank_zp_idx_list.append(pbunch)
    return pbunch


def ranking_reduction_trials(
    aid,
    rank_zp_idx_list,
    num_total_max=4,
    num_type_max=2,
    num_try2=2,
    ranking_factor_max=None,
    greedy=False,
    return_remaining=False,
    reset_delay=True,
    hint_name=None,
):
    aid = ensure_aid(aid)
    if not rank_zp_idx_list:
        if return_remaining:
            return [], rank_zp_idx_list
        else:
            return []

    type_count = dict()
    num_total = 0
    idx_total = 0

    trials = []
    trials_try1 = []
    trials_try2 = []
    ranks_original = []
    minranking = float("inf")

    aid.log_progress(6, "trials started")
    for pbunch in rank_zp_idx_list:
        idx_total += 1

        if ranking_factor_max is not None and (
            pbunch.rank > minranking * ranking_factor_max
        ):
            continue
        if pbunch.rank < minranking:
            minranking = pbunch.rank

        count = type_count.get(pbunch.name, 0) + 1
        if num_type_max is not None and count > num_type_max:
            continue
        type_count[pbunch.name] = count
        num_total += 1
        if num_total_max is not None and num_total > num_total_max:
            break
        ranks_original.append(pbunch.rank)

        Pc_mod = list(aid.fitter.poles.c)
        Zc_mod = list(aid.fitter.zeros.c)
        Pr_mod = list(aid.fitter.poles.r)
        Zr_mod = list(aid.fitter.zeros.r)
        Pc_idx_remove = set()
        Zc_idx_remove = set()
        Pr_idx_remove = set()
        Zr_idx_remove = set()
        Pc_new = []
        Zc_new = []
        Pr_new = []
        Zr_new = []
        # work backward on the command sequence removing the idxs
        trial = Bunch()
        trial.prog_redux = []
        trial.ord_ch = 0
        trial.pbunch = pbunch
        program = list(pbunch.program)
        while program:
            arg = program.pop()
            which = program.pop()
            trial.prog_redux.append(which)
            if which == "PcDelIdx":
                trial.prog_redux.append(Pc_mod[arg])
                trial.ord_ch -= 2
                Pc_idx_remove.add(arg)
            elif which == "ZcDelIdx":
                trial.prog_redux.append(Zc_mod[arg])
                trial.ord_ch -= 2
                Zc_idx_remove.add(arg)
            elif which == "PrDelIdx":
                trial.prog_redux.append(Pr_mod[arg])
                trial.ord_ch -= 1
                Pr_idx_remove.add(arg)
            elif which == "ZrDelIdx":
                trial.prog_redux.append(Zr_mod[arg])
                trial.ord_ch -= 1
                Zr_idx_remove.add(arg)
            elif which == "PrAdd":
                trial.prog_redux.append(arg)
                trial.ord_ch += 1
                Pr_new.append(arg)
            elif which == "ZrAdd":
                trial.prog_redux.append(arg)
                trial.ord_ch += 1
                Zr_new.append(arg)
            elif which == "PcAdd":
                trial.prog_redux.append(arg)
                trial.ord_ch += 2
                Pc_new.append(arg)
            elif which == "ZcAdd":
                trial.prog_redux.append(arg)
                trial.ord_ch += 2
                Zc_new.append(arg)
        for idx in reversed(sorted(Pc_idx_remove)):
            del Pc_mod[idx]
        for idx in reversed(sorted(Zc_idx_remove)):
            del Zc_mod[idx]
        for idx in reversed(sorted(Pr_idx_remove)):
            del Pr_mod[idx]
        for idx in reversed(sorted(Zr_idx_remove)):
            del Zr_mod[idx]

        if trial.ord_ch == 0:
            trial.ord_str = "OrdC"
        elif trial.ord_ch < 0:
            trial.ord_str = "OrdDn"
        else:
            trial.ord_str = "OrdUp"

        coding_map, num_codings_mod, den_codings_mod = ZP2codings(
            aid.fitter,
            zeros=representations.asMRRB(r=Zr_mod, c=Zc_mod),
            poles=representations.asMRRB(r=Pr_mod, c=Pc_mod),
        )
        coding_map, num_codings_new, den_codings_new = ZP2codings(
            aid.fitter,
            zeros=representations.asMRRB(r=Zr_new, c=Zc_new),
            poles=representations.asMRRB(r=Pr_new, c=Pc_new),
            coding_map=coding_map,
        )
        fitter = coding_map.mrf_default(
            parent=aid.fitter,
            num_codings=num_codings_mod + num_codings_new,
            den_codings=den_codings_mod + den_codings_new,
        )

        trial.fitter = fitter
        trial.codings_mod = num_codings_mod + den_codings_new
        trial.codings_new = num_codings_mod + den_codings_new
        trial.rank_original = pbunch.rank
        trials_try1.append(trial)

    def trial_optimize(trial):
        improved = False
        fitter = trial.fitter
        try:
            with fitter.with_codings_only([fitter.gain_coding]):
                fitter.optimize()
        except Exception as e:
            aid.log_debug(9, "Optimize Exception", e)

        try:
            # anneal by moving the original codings first
            #if trial.codings_new:
            #    with fitter.with_codings_only(trial.codings_mod):
            #        fitter.optimize()
            if reset_delay:
                aid.fitter.delay_s = aid.hint("delay_s")
            #fitter.optimize_NM()
            fitter.optimize()
        except Exception as e:
            aid.log_debug(9, "Optimize Exception", e)
            trial.improved = False
            return trial

        if trial.pbunch.variant is not None:
            ord_str = trial.pbunch.variant
        else:
            ord_str = trial.ord_str

        improved = aid.fitter_check(
            fitter,
            variant=ord_str,
            update=False,
            validate=False,
            hint_name=hint_name,
        )
        trial.improved = improved
        return trial

    mt = aid.hint("multithreading", None)
    if mt is not None and mt > 1:
        import multiprocessing.pool

        pool = multiprocessing.pool.ThreadPool(
            processes=mt,
        )
        trial_map = pool.imap_unordered
    else:
        trial_map = map

    improved = False
    for trial in trial_map(trial_optimize, trials_try1):
        aid.log(8, "Reducing Program", pbunch.rank, trial.prog_redux)
        if not trial.improved:
            trials_try2.append(trial)
        else:
            improved = True
            trials_try2.append(trial)
        trials.append(trial)
        if greedy and trial.improved:
            break

    aid.log_progress(7, "trials annealing")

    def trial_optimize_anneal(trial):
        fitter = trial.fitter
        # TODO fix
        from . import algorithms

        if not algorithms.optimize_anneal(aid, fitter):
            return trial
        improved = aid.fitter_check(
            fitter,
            variant=trial.ord_str,
            update=False,
            validate=False,
        )
        trial.improved = improved
        return trial

    if not (greedy and improved):
        trials_try2.sort(key=lambda trial: trial.fitter.residuals_average)
        for trial in trial_map(trial_optimize_anneal, trials_try2[:num_try2]):
            if greedy and trial.improved:
                break

    trials.sort(key=lambda trial: trial.fitter.residuals_average)
    if not return_remaining:
        return trials
    else:
        return trials, rank_zp_idx_list[idx_total:]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np


from .MRFprogram import (
    resrank_program,
    ranking_reduction_trials,
)


def sign_check_flip(fitter, max_N=None):
    """ """
    if max_N is None:
        xfer = fitter.xfer_fit
        data = fitter.data
        W = fitter.W
    else:
        xfer = fitter.xfer_fit[:max_N]
        data = fitter.data[:max_N]
        W = fitter.W[:max_N]
    rat = data / xfer
    select = np.isfinite(rat)
    rat_ang = np.exp(1j * np.angle(rat))
    ang_avg_fit = np.sum(rat_ang[select] * W[select] ** 2) / np.sum(W[select] ** 2)

    residuals = fitter.residuals_average
    if ang_avg_fit.real < 0:
        fitter.gain = -fitter.gain

    if fitter.residuals_average > residuals:
        fitter.gain = -fitter.gain

    if max_N is None:
        xfer = fitter.xfer_fit
        data = fitter.data
        W = fitter.W
    else:
        xfer = fitter.xfer_fit[:max_N]
        data = fitter.data[:max_N]
        W = fitter.W[:max_N]
    rat = data / xfer
    rat_ang = np.exp(1j * np.angle(rat))
    ang_avg_fit = np.sum(rat_ang[select] * W[select] ** 2) / np.sum(W[select] ** 2)
    # assert ang_avg_fit.real > 0
    return


def optimize_anneal(aid, fitter=None):
    if fitter is None:
        fitter = aid.fitter
    try:
        # TODO, maybe try nelder_mead here
        fitter.optimize(
            residuals_type=aid.hint("residuals_type_alt", "log"),
            aid=aid,
            max_nfev=30,
        )
    except Exception as e:
        aid.log_debug(10, "Optimize Exception, ", e)

    try:
        fitter.optimize(aid=aid)
    except Exception as e:
        aid.log_debug(10, "Optimize Exception, ", e)
        return False
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


def ZPKrep_DF2_noise(
    ZPK,
    RMSin=1,
    ZPKin=[
        (),
        (
            -0.1,
            -0.1,
        ),
        1,
    ],
    N_quant=3e-16,
    F_nyquist=None,
):
    """
    From DCC G0900928 (M. Evans noise report)
    """
    return


def ZPKrep_BQ_noise(
    ZPK,
    RMSin=1,
    ZPKin=[
        (),
        (
            -0.1,
            -0.1,
        ),
        1,
    ],
    N_quant=3e-16,
    F_nyquist=None,
):
    """
    From DCC G0900928 (M. Evans noise report)
    """
    8 * N_quant * np.max(RMS_in, RMS_out)
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from wield.iirrational.testing.utilities import (
    sign_validate_and_plot_hint,
    sign_validate_and_digest_hint,
)

from ..testing.plots import plot_on_assert


def validate_plot_log(fname, request):
    hdict = sign_validate_and_plot_hint(__file__, request)
    hdict.update(
        dict(
            log_level_debug=10,
            log_level=10,
            log_level_warn=10,
        )
    )
    return hdict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from .data2filter import data2filter
from .disc_sequence import rational_disc_fit, ratdisc_single
from .disc_sequence_mag import rational_disc_fit_mag
from .fit_aid import FitAid
from ..data2testcase import (
    data2testcase,
    testcase2data,
)
from . import hintsets



__all__ = [
    "data2filter",
    "data2testcase",
    "rational_disc_fit,"
    "rational_disc_fit_mag",
    "ratdisc_single,"
    "FitAid",
    "hintsets,"
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np

from ..TFmath.statespace import ss2zpk, ss2xfer
from .. import fitters_ZPK
from . import data2filter


def ss2filter(
    A,
    B,
    C,
    D,
    F_Hz,
    SNR=1e3,
    idx_in=None,
    idx_out=None,
    mode="reduce",
    prune_Qrank=0.01,
    prune_Qrank_unstable=0.5,
    **kwargs
):
    z, p, k = ss2zpk(
        A,
        B,
        C,
        D,
        idx_in=idx_in,
        idx_out=idx_out,
        fmt="IIRrational",
        Q_rank_cutoff_unstable=prune_Qrank_unstable,
    )

    fit = data2filter(
        data=ss2xfer(A, B, C, D, F_Hz, idx_in=idx_in, idx_out=idx_out),
        F_Hz=F_Hz,
        zeros=z,
        poles=p,
        gain=k,
        SNR=SNR,
        mode=mode,
        prune_Qrank=prune_Qrank,
        coding_map=fitters_ZPK.coding_maps.RI,
        **kwargs
    )
    return fit



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


# import numpy as np

from ..codings_cmn import (
    CodingType,
    # Ipi,
    # I2pi
)

# import scipy.linalg


class CodingRI(CodingType):
    N_parameters = 2
    p_real = 0
    p_imag = 0
    lock_real = False
    lock_imag = False
    hide_real = False
    hide_imag = False

    def setup(
        self,
        hide_all=None,
        lock_real=None,
        hide_real=None,
        lock_imag=None,
        hide_imag=None,
        disable=None,
    ):
        if hide_all:
            hide_real = True
            hide_imag = True

        N_parameters = 2
        if lock_real is not None:
            self.lock_real = lock_real
        if hide_real is not None:
            self.hide_real = hide_real
        if self.hide_real:
            N_parameters -= 1

        if lock_imag is not None:
            self.lock_imag = lock_imag
        if hide_imag is not None:
            self.hide_imag = hide_imag
        if self.hide_imag:
            N_parameters -= 1

        if disable is not None:
            self.disable = disable
        if self.disable:
            N_parameters = 0

        self.N_parameters = N_parameters

    def update(self, A=None, B=None):
        if self.disable:
            assert A is None and B is None
            return
        if self.hide_imag:
            assert B is None
            if self.hide_real:
                assert A is None
            else:
                real = A
                if not self.lock_real:
                    self.p_real = real
        else:
            if self.hide_real:

                imag = A
                if not self.lock_imag:
                    self.p_imag = imag
            else:
                real, imag = A, B
                if not self.lock_real:
                    self.p_real = real

                if not self.lock_imag:
                    self.p_imag = imag

    def reduce(self):
        if self.disable:
            return []
        if self.hide_imag:
            if self.hide_real:
                return []
            else:
                return [self.p_imag]
        else:
            if self.hide_real:
                return [self.p_imag]
            else:
                return [
                    self.p_real,
                    self.p_imag,
                ]

    def update_roots(self, r1):
        """
        r2, may be unspecified, in which case it is assumed to be nothing, if r1 is real, or otherwise the conjugate of r1
        """
        self.p_real = r1.real
        self.p_imag = r1.imag
        return

    def transfer(self):
        # real/imaginary part of root
        r, i = self.p_real, self.p_imag
        X = self.sys.Xsf_grid
        Xsq = self.sys.Xsf_grid_sq
        return (r * r + i * i) - 2 * X * r + Xsq

    def derivative(self):
        if self.disable:
            return []
        # real/imaginary part of root
        r, i = self.p_real, self.p_imag
        X = self.sys.Xsf_grid
        if not self.hide_real:
            if not self.hide_imag:
                return [
                    2 * (r - X) if not self.lock_real else 0,
                    (2 * i) if not self.lock_imag else 0,
                ]
            else:
                return [
                    2 * (r - X) if not self.lock_real else 0,
                ]
        else:
            if not self.hide_imag:
                return [
                    (2 * i) if not self.lock_imag else 0,
                ]
            else:
                return []

    def roots_c(self):
        # real/imaginary part of root
        return [self.p_real + 1j * abs(self.p_imag)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np

from ..root_bunch import root_constraints, RBAlgorithms

from .constraints import poly_constraints

fromroots = np.polynomial.polynomial.polyfromroots
val = np.polynomial.polynomial.polyval
vander = np.polynomial.polynomial.polyvander
RBalgo = RBAlgorithms(strict=False)


def roots_rB(c, constraint):
    if poly_constraints.even_real <= constraint:
        assert np.all(c[::2].imag == 0)
    if poly_constraints.odd_real <= constraint:
        assert np.all(c[1::2].imag == 0)
    if poly_constraints.odd_imag <= constraint:
        assert np.all(c[1::2].real == 0)

    if poly_constraints.no_constraint == constraint:
        rvec = roots(c)
        rB = RBalgo.expect(rvec, constraint=root_constraints.no_constraint)
    elif poly_constraints.eRoR == constraint:
        rvec = roots(c)
        rB = RBalgo.expect(rvec, constraint=root_constraints.mirror_real)
    elif poly_constraints.eRoI == constraint:
        rvec = roots(c)
        rB = RBalgo.expect(rvec, constraint=root_constraints.mirror_imag)
    elif poly_constraints.eRoZ == constraint:
        rvec = roots(c)
        rB = RBalgo.expect(rvec, constraint=root_constraints.mirror_quad)
    return rB


def roots(c):
    return np.polynomial.polynomial.polyroots(c)


def coeff_canonicalization_gain(c):
    return c[-1]


def roots_lnG(c):
    lnG = np.log(c[-1])
    return np.polynomial.polynomial.roots(c), lnG


def fromroots_lnG(roots):
    roots = np.asarray(roots)
    return fromroots(roots), 0


def val_lnG(X, c, lnG=0):
    return val(X, c), lnG


def vander_lnG(X, N, lnG=0):
    return vander(X, N), lnG



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from wield.bunch import Bunch

from ..utilities import ensure_aid
from .. import TFmath
from .. import representations


def nyquist_move(
    fitter,
    nyquist_new,
    clear_neg_real=False,
    split_neg_real=False,
    aid=None,
):
    """
    Modifies a fitter IN PLACE to use a different nyquist frequency
    """
    # TODO, update this to properly use RootBunches and ZPKreps
    aid = ensure_aid(aid)
    ZPKrep = fitter.ZPKrep

    F_nyquist_Hz = ZPKrep.F_nyquist_Hz
    op = Bunch(gain_flip=1)

    # currently a restriction in the algorithm, it can't handle aliasing of poles (need bilinear for that)
    assert nyquist_new > F_nyquist_Hz
    ratio = nyquist_new / F_nyquist_Hz

    def root_list_map(root_list, rtype):
        new_list = []
        for root in root_list:
            new_roots = None
            if abs(root.imag) < 1e-15:
                # aid.log(root)
                if root.real > 0:
                    if root.real < 1:
                        new_roots = [1 - (1 - root.real) / ratio]
                    else:
                        # actually, just clear it
                        aid.log("Cleared ", rtype, root)
                        # new_roots = []
                        # flip it then insert
                        op.gain_flip *= -1
                        root = 1 / root
                        # new_roots = [1 - (1 - root.real) / ratio]
                        new_roots = []
                else:
                    if clear_neg_real:
                        aid.log("Cleared ", rtype, root)
                        new_roots = []
                    elif split_neg_real:
                        angle = np.pi / ratio
                        amp = -root.real
                        if amp < 1:
                            amp = 1 - (1 - amp) / ratio
                        else:
                            amp = (amp - 1) / ratio + 1
                        new_roots = [
                            amp * np.exp(1j * angle),
                            amp * np.exp(-1j * angle),
                        ]
                    else:
                        new_roots = [-1 - (-1 - root.real) / ratio]
            else:
                amp = abs(root)
                angle = np.angle(root)
                if angle > 0:
                    angle = angle / ratio
                else:
                    angle = angle / ratio
                if amp < 1:
                    amp = 1 - (1 - amp) / ratio
                else:
                    amp = (amp - 1) / ratio + 1
                new_roots = [amp * np.exp(1j * angle)]
            new_list.extend(new_roots)
        return new_list

    xfer_pre = fitter.xfer_fit

    nyquist_prev = ZPKrep.F_nyquist_Hz
    fitter.F_nyquist_Hz = nyquist_new
    poles = root_list_map(fitter.poles.fullplane, rtype="pole")
    zeros = root_list_map(fitter.zeros.fullplane, rtype="zero")
    fitter.poles = poles
    fitter.zeros = zeros

    # don't use the root mapper above since it does special things and can drop
    # overlays
    Zov, Pov, Kex = TFmath.SorZtoSorZ(
        (fitter.zeros_overlay.fullplane, fitter.poles_overlay.fullplane, 1),
        F_nyquist_Hz_in=nyquist_prev,
        F_nyquist_Hz_out=nyquist_new,
    )
    fitter.poles_overlay = Pov
    fitter.zeros_overlay = Zov

    xfer_post = fitter.xfer_fit

    gain_scale = np.median(abs(xfer_post / xfer_pre))

    gain = fitter.gain / gain_scale * op.gain_flip
    fitter.gain = gain
    # aid.log("GAIN: ", fitter.gain)
    return fitter


def nyquist_remove(
    fitter,
    clear_neg_real=False,
    split_neg_real=False,
    aid=None,
):
    """
    Modifies a fitter IN PLACE to use a different nyquist frequency
    """
    aid = ensure_aid(aid)

    F_nyquist_Hz = fitter.F_nyquist_Hz
    op = Bunch(gain_flip=1)

    def root_list_map(root_list, rtype):
        new_list = []
        for root in root_list:
            new_roots = None
            if abs(root.imag) < 1e-15:
                # aid.log(root)
                if root.real > 0:
                    if root.real < 1:
                        new_roots = [F_nyquist_Hz * (root.real - 1)]
                    else:
                        aid.log("Cleared ", rtype, root)
                        op.gain_flip *= -1
                        new_roots = []
                else:
                    if clear_neg_real:
                        aid.log("Cleared ", rtype, root)
                        # TODO, not sure about this one
                        op.gain_flip *= -1
                        new_roots = []
                    elif split_neg_real:
                        BW = -(1 + root.real) * F_nyquist_Hz
                        new_roots = [
                            BW + 1j * F_nyquist_Hz,
                            BW + -1j * F_nyquist_Hz,
                        ]
                    else:
                        new_roots = [F_nyquist_Hz * -(root.real + 1)]
            else:
                amp = abs(root)
                F_Hz = np.angle(root) / np.pi * F_nyquist_Hz
                new_roots = [(amp - 1) * F_nyquist_Hz + 1j * F_Hz]
            new_list.extend(new_roots)
        return new_list

    xfer_pre = fitter.xfer_fit

    poles = root_list_map(fitter.poles.fullplane, rtype="pole")
    zeros = root_list_map(fitter.zeros.fullplane, rtype="zero")

    # don't use the root mapper above since it does special things and can drop
    # overlays
    Zov, Pov, Kex = TFmath.ZtoS(
        (fitter.zeros_overlay.fullplane, fitter.poles_overlay.fullplane, 1),
        F_nyquist_Hz=fitter.F_nyquist_Hz,
    )

    ZPKrep = representations.ZPKwData(
        parent=fitter.ZPKrep,
        poles=poles,
        zeros=zeros,
        poles_overlay=Zov,
        zeros_overlay=Pov,
        gain=Kex * fitter.ZPKrep.gain,
    )

    xfer_post = ZPKrep.xfer_fit

    gain_scale = np.median(abs(xfer_post / xfer_pre))
    gain = fitter.gain / gain_scale * op.gain_flip

    ZPKrep.gain = gain
    return ZPKrep



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from matplotlib import gridspec

from ..utilities.mpl import generate_stacked_plot_ax
from .plot_augment import plot_augment_residuals


def plot_residuals(
    self,
    fitter=None,
    gs_base=gridspec.GridSpec(1, 1)[0],
    fig=None,
    ax_bunch=None,
    xscale=None,
    plot_now=True,
    **kwargs
):
    use_phase = True

    if ax_bunch is None:
        axB = generate_stacked_plot_ax(
            [
                ("mag", True),
                ("phase", use_phase),
            ],
            height_ratios=dict(
                mag=2,
            ),
            heights_phys_in_default=1,
            xscales="log",
            gs_base=gs_base,
            fig=fig,
            ax_bunch=ax_bunch,
        )
    else:
        axB = ax_bunch

    fit_list = []
    axB.plot_fit = plot_augment_residuals(
        self, ax_mag=axB.mag, ax_phase=axB.phase, fit_list=fit_list, **kwargs
    )

    axB.mag.set_ylabel("Residuals Magnitude")
    if axB.phase:
        axB.phase.set_ylabel("Residuals Phase [deg]")
    axB.ax_bottom.set_xlabel("Frequency [Hz]")

    if fitter is not None and plot_now:
        axB.plot_fit(fitter)

    def finalize():
        min_lims = []
        max_lims = []
        for fitB in fit_list:
            min_err_lim = min(
                fitB.median_res_lim,
                fitB.min_res_err_lim,
                max(fitB.min_res_err_lim / 30, fitB.min_res_err_lim * 1),
            )
            min_lims.append(min_err_lim)
            max_err_lim = max(
                fitB.median_res_lim,
                fitB.max_res_err_lim,
                min(fitB.max_res_err_lim * 30, fitB.max_res_err_lim / 1),
            )
            max_lims.append(max_err_lim)
        axB.mag.set_ylim(np.min(min_lims), max(max_lims))

        scales = []
        mins = []
        minsNZ = []
        maxs = []
        for fitB in fit_list:
            scales.append(fitB.scale_log)
            mins.append(fitB.minF_Hz)
            minsNZ.append(fitB.minF_HzNZ)
            maxs.append(fitB.maxF_Hz)

        if xscale is None:
            if np.count_nonzero(scales) / len(scales) > 0.5:
                final_xscale = "log_zoom"
            else:
                final_xscale = "linear"
        else:
            final_xscale = xscale

        if final_xscale in ["log", "log_zoom"]:
            xmin = np.min(minsNZ)
        else:
            xmin = np.min(mins)
        xmax = np.max(maxs)
        axB.ax_bottom.set_xscale(final_xscale)
        axB.ax_bottom.set_xlim(xmin, xmax)
        for subax in axB.ax_list_0:
            subax.set_xscale(final_xscale)
            subax.set_xlim(xmin, xmax)
        axB.mag.legend(ncol=2, fontsize=8, loc="best")

    axB.finalizers.append(finalize)
    return axB



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
import scipy

# from wield import declarative
import scipy.linalg
import scipy.special

from ..svd import SVD_SV
from .. import TFmath
from ..representations.polynomials import poly_constraints
from .rational_bases import DataFilterBase


def lstsq_safe(a, b, *args, **kwargs):
    """
    On some machines, LAPACK can fail on the gelsd driver call on some fits
    The other drivers are more resilient
    """
    driver = kwargs.get("lapack_driver", None)
    if driver is None:
        kwargs["lapack_driver"] = "gelsd"
    try:
        return scipy.linalg.lstsq(a, b, *args, **kwargs)
    except Exception as E:
        import warnings

        warnings.warn("LAPACK lstsq error: {}\ntrying driver gelss".format(E))

    try:
        kwargs["lapack_driver"] = "gelss"
        return scipy.linalg.lstsq(a, b, *args, **kwargs)
    except Exception as E:
        import warnings

        warnings.warn("LAPACK lstsq error: {}\ntrying driver gelsy".format(E))

    kwargs["lapack_driver"] = "gelsy"
    return scipy.linalg.lstsq(a, b, *args, **kwargs)


class PolyFilterAlgorithmsBase(DataFilterBase):
    def match_pair_iter(
        self,
        Q_rank_cutoff,
        num_sequence=2,
        zeros_first=True,
    ):
        npoles = self.npoles
        nzeros = self.nzeros

        self.matched_pairs_clear(Q_rank_cutoff=Q_rank_cutoff)

        self.npoles = len(self.poles)
        self.nzeros = len(self.zeros)

        if not zeros_first:
            self.fit_poles()
            num_sequence -= 1
        for idx in range(num_sequence):
            self.fit_zeros()
            self.fit_poles()

        if not zeros_first:
            self.fit_zeros()

        self.npoles = npoles
        self.nzeros = nzeros

    def matched_pairs_clear(
        self,
        Q_rank_cutoff=0.5,
    ):
        """
        Match unique closest pairs, if they are within a bandwidth of 0Hz, then they are ignored
        """
        if Q_rank_cutoff == 0:
            return

        # cast into native representation
        poles = self.phi_rep_Snative(self.poles)
        zeros = self.phi_rep_Snative(self.zeros)

        # TODO, not working for rational_disc since it add branch-cut info
        # poles = self.RBalgo.expect(poles, self.RBalgo.root_constraints.mirror_real)
        # zeros = self.RBalgo.expect(zeros, self.RBalgo.root_constraints.mirror_real)

        rpB = TFmath.nearest_unique_pairs(zeros.c, poles.c)

        rpB.r12_list
        rpB.l1_remain
        rpB.l2_remain

        poles_new = []
        zeros_new = []
        for z, p in rpB.r12_list:
            Q_rank = self.Q_rank_Snative(z, p)
            if Q_rank > Q_rank_cutoff:
                poles_new.append(p)
                zeros_new.append(z)

        poles.c = np.concatenate([poles_new, rpB.l2_remain])
        zeros.c = np.concatenate([zeros_new, rpB.l1_remain])

        self.zeros = self.phi_Snative_rep(zeros)
        self.poles = self.phi_Snative_rep(poles)
        return

    def root_stabilize_Snative_MIP(
        self,
        rB,
        method=None,
    ):
        """
        MIP stands for modify-in-place, which must occur for the Snative
        representation
        """
        if method is None:
            return rB

        select_c = rB.c.real > 0
        select_r = rB.r.real > 0
        if method == "flip":
            rB.r[select_r] = -rB.r[select_r]
            rB.c[select_c] = -rB.c[select_c].conjugate()
        elif method == "remove":
            rB.r = rB.r[~select_r]
            rB.c = rB.c[~select_c]
        else:
            raise RuntimeError("Bad Argument")

        rB.clear()
        return

    def root_stabilize(self, rB, method=None):
        rB_native = self.phi_rep_Snative(rB)
        self.root_stabilize_Snative_MIP(rB_native, method=method)
        return self.phi_Snative_rep(rB_native)

    def Q_rank_Snative(self, p, z):
        z_soft = self.root_BW_soften_Snative(z)
        p_soft = self.root_BW_soften_Snative(p)
        return (
            abs(p_soft - z_soft) * (1 / p_soft.real ** 2 + 1 / z_soft.real ** 2) ** 0.5
        )

    def root_BW_soften_Snative(self, root):
        # TODO, convert this into a distance_limit type function, to use the same
        # as ZPK_fitters
        F = root.imag
        F_abs = abs(F)
        BW = root.real
        BW_abs = abs(BW)

        if F_abs > self.F_max_Hz:
            dist = F_abs - self.F_max_Hz
            if BW_abs < dist / 2:
                BW = np.sign(BW) * dist / 2
        else:
            idx_nearest = np.searchsorted(self.F_Hz, F_abs)
            idx_min = max(idx_nearest - 1, 0)
            idx_max = min(idx_nearest + 2, len(self.F_Hz) - 1)
            BW_near = self.F_Hz[idx_max] - self.F_Hz[idx_min]
            if BW_abs < BW_near:
                BW = np.sign(BW) * BW_near
        return BW + 1j * F


class PolyFilterAlgorithms(PolyFilterAlgorithmsBase):
    """
    Mixin Class
    """

    poly_constraints = poly_constraints.eRoR

    def poly_fromroots(self, roots):
        """
        Cleans the coefficients as well
        """
        pvec, lnG = self.poly.fromroots_lnG(roots, X_scale=self.X_scale)
        return pvec.real

    def fit_poles_mod_zeros(self, **kwargs):
        # remove the effect of linear (delay) phasing
        A_z = self.A_z
        # A_z = np.hstack([(1j*F_Hz).reshape(-1, 1), A_z])
        q, r = np.linalg.qr(A_z)

        A_p = self.A_p
        # the str is because py2 doesn't like some symbol in there
        A_p = A_p - np.einsum(
            str("ij,jk->ik"), q, np.einsum(str("ij,ik->jk"), q.conjugate(), A_p)
        )

        S, V = SVD_SV(
            np.vstack([A_p.real, A_p.imag]),
            n_smallest=3,
            overwrite_a=True,
        )
        # print("POLES SVD: ", S[-4:])
        self.a_vec = V.T[:, -1]
        return

    def fit_SVD(self, modify=True, residuals=0, **kwargs):
        A_z = self.A_z
        A_p = self.A_p
        if residuals == 0:
            pass
        elif residuals > 0:
            A_z = self.V_b * (self.W / (self.h_b / self.h_delay)).reshape(-1, 1)
        elif residuals < 0:
            A_p = self.V_a * (self.W / (self.h_delay * self.h_a)).reshape(-1, 1)

        A_svd = np.hstack(
            [np.vstack([A_p.real, A_p.imag]), -np.vstack([A_z.real, A_z.imag])]
        )

        S, V = SVD_SV(
            A_svd,
            n_smallest=3,
            overwrite_a=True,
        )

        # ignore division by zero warnings since NaN output is correct.
        # Argmin does the right thing with nan
        with np.errstate(divide="ignore", invalid="ignore"):
            S_rescaled = abs(S / V.T[0, :])

        idx_best = np.argmin(S_rescaled)
        if modify:
            self.a_vec = V.T[: A_p.shape[1], idx_best]
            self.b_vec = V.T[A_p.shape[1] :, idx_best]
        return S, V.T[0, :]

    def fit_zeros_mod_poles(self, **kwargs):
        A_p = self.A_p
        # remove the effect of linear (delay) phasing
        # A_p = np.hstack([(1j*F_Hz).reshape(-1, 1), A_p])

        q, r = np.linalg.qr(A_p)

        A_z = self.A_z
        A_z = A_z - np.einsum(
            str("ij,jk->ik"), q, np.einsum(str("ij,ik->jk"), q.conjugate(), A_z)
        )
        S, V = SVD_SV(
            np.vstack([A_z.real, A_z.imag]),
            n_smallest=3,
            overwrite_a=True,
        )
        # print("ZEROS SVD: ", S[-4:])
        self.b_vec = V.T[:, -1]
        return

    def fit_poles(self, **kwargs):
        # print(self.a_vec, self.b_vec)
        A_p = self.A_p
        # solve the problem with purely real taps
        a_fit, res, rank, s = lstsq_safe(
            np.vstack([A_p.real, A_p.imag]),
            np.hstack([self.W.real, self.W.imag]),
        )
        # print(a_fit)
        self.a_vec = a_fit
        return

    def fit_poles2(self, **kwargs):
        A_p = self.A_p
        # print(A_p.real.shape)
        a_fit, res, rank, s = lstsq_safe(
            np.block(
                [
                    [A_p.real, 0 * self.W.reshape(-1, 1)],
                    [A_p.imag, (self.W * self.F_Hz).reshape(-1, 1)],
                ]
            ),
            np.hstack([self.W.real, self.W.imag]),
        )
        self.a_vec = a_fit[:-1]
        return

    def fit_zeros(self, **kwargs):
        A_z = self.A_z
        b_fit, res, rank, s = lstsq_safe(
            np.vstack([A_z.real, A_z.imag]),
            np.hstack([self.W.real, self.W.imag]),
        )
        self.b_vec = b_fit
        return

    def fit_zeros2(self, **kwargs):
        A_z = self.A_z
        b_fit, res, rank, s = lstsq_safe(
            np.block(
                [
                    [A_z.real, 0 * self.W.reshape(-1, 1)],
                    [A_z.imag, (self.W * self.F_Hz).reshape(-1, 1)],
                ]
            ),
            np.hstack([self.W.real, self.W.imag]),
        )
        self.b_vec = b_fit[:-1]
        return


class PolyFilterAlgorithmsIm(PolyFilterAlgorithmsBase):
    """
    Mixin Class for polynomial algorithms which apply the constraint to mirror
    over the imaginary axis rather than the real.
    """

    poly_constraints = poly_constraints.eRoI

    def poly_fromroots(self, roots):
        """
        Cleans the coefficients as well
        """
        # print('ROOTS: ', roots)
        pvec, lnG = self.poly.fromroots_lnG(roots, X_scale=self.X_scale)
        if np.iscomplexobj(pvec):
            pvec[0::2].imag = 0
            pvec[1::2].real = 0
        return pvec

    # TODO
    def fit_SVD(self, modify=True, residuals=0, **kwargs):
        coeff_mult = np.empty(self.A_p.shape[1], dtype=complex)
        coeff_mult[0::2] = 1
        coeff_mult[1::2] = 1j
        A_p = coeff_mult.reshape(1, -1) * self.A_p

        coeff_mult = np.empty(self.A_z.shape[1], dtype=complex)
        coeff_mult[0::2] = 1
        coeff_mult[1::2] = 1j
        A_z = coeff_mult.reshape(1, -1) * self.A_z

        if residuals == 0:
            pass
        elif residuals > 0:
            A_z = self.V_b * (self.W / (self.h_b / self.h_delay)).reshape(-1, 1)
        elif residuals < 0:
            A_p = self.V_a * (self.W / (self.h_delay * self.h_a)).reshape(-1, 1)

        A_svd = np.hstack(
            [
                np.vstack([A_p.real, A_p.imag]),
                -np.vstack([A_z.real, A_z.imag]),
            ]
        )

        S, V = SVD_SV(
            A_svd,
            n_smallest=3,
            overwrite_a=True,
        )
        with np.errstate(divide="ignore", invalid="ignore"):
            S_rescaled = abs(S / V.T[0, :])
        idx_best = np.argmin(S_rescaled)

        a_fit = V.T[: A_p.shape[1], idx_best]
        b_fit = V.T[A_p.shape[1] :, idx_best]

        a_fit = a_fit.astype(complex, copy=True)
        b_fit = b_fit.astype(complex, copy=True)

        a_fit[1::2] *= 1j
        b_fit[1::2] *= 1j

        if modify:
            self.a_vec = a_fit
            self.b_vec = b_fit
        return S, a_fit, b_fit

    # TODO
    def fit_poles(self, **kwargs):
        # print(self.a_vec, self.b_vec)
        coeff_mult = np.empty(self.A_p.shape[1], dtype=complex)
        coeff_mult[0::2] = 1
        coeff_mult[1::2] = 1j

        A_p = coeff_mult.reshape(1, -1) * self.A_p
        # solve the problem with purely real taps
        a_fit, res, rank, s = lstsq_safe(
            np.vstack([A_p.real, A_p.imag]),
            np.hstack([self.W.real, self.W.imag]),
        )
        # print("Sp: ", s / s[0])
        a_fit = a_fit.astype(complex, copy=True)
        a_fit[1::2] *= 1j
        self.a_vec = a_fit
        return

    # TODO
    def fit_zeros(self, **kwargs):
        coeff_mult = np.empty(self.A_z.shape[1], dtype=complex)
        coeff_mult[0::2] = 1
        coeff_mult[1::2] = 1j

        A_z = coeff_mult.reshape(1, -1) * self.A_z
        b_fit, res, rank, s = lstsq_safe(
            np.vstack([A_z.real, A_z.imag]),
            np.hstack([self.W.real, self.W.imag]),
        )
        # print("Sz: ", s / s[0])
        b_fit = b_fit.astype(complex, copy=True)
        b_fit[1::2] *= 1j
        self.b_vec = b_fit
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np


def load_csv(fname, parse_str, parse_map):
    """
    Can raise KeyError if parse_str does not map into parse_map
    """
    delimiter = parse_str[0]
    if delimiter in ["-"]:
        delimiter = None
    parse_str = parse_str[1:]

    if delimiter is None:
        subkeyspl = parse_str.split()
    else:
        subkeyspl = parse_str.split(delimiter)

    key_gen = []
    for sk in subkeyspl:
        key_gen.append(parse_map[sk])

    farr = np.genfromtxt(
        fname,
        delimiter=delimiter,
        filling_values=float("NaN"),
    )

    fdict = dict()
    for idx_col in range(farr.shape[0]):
        fdict[str(idx_col)] = farr[idx_col]

        if idx_col < len(subkeyspl):
            sk = key_gen[idx_col]
            if sk is not None:
                fdict[sk] = farr[idx_col]

    return fdict



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest
from os import path

from wield.iirrational import v2
from wield.iirrational.testing.plots import plot_on_assert

from wield.iirrational.testing.utilities import (
    sign_validate_and_plot_hint,
    stability_validate_and_plot_hint,
)

try:
    from IIRrational_test_data import IIRrational_data_dev
    from IIRrational_test_data import matlab as matlab_test_data
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False

localpath = path.split(__file__)[0]
skipdeco = pytest.mark.skipif(
    module_import_skip, reason="cannot import IIRrational_test_data"
)


@skipdeco
def test_oplev_iy(request, browser, plotsections, plot_verbosity):
    from wield.iirrational.file_io import load_any

    fname = matlab_test_data.matfiles["OplevPlant.mat"]
    fdict = load_any(fname=fname, ftype="mat")

    out = v2.data2filter(
        F_Hz=fdict["ap"]["iy"]["ff"],
        data=fdict["ap"]["iy"]["plant"],
        F_nyquist_Hz=None,
        delay_s=0,
        hints=[sign_validate_and_plot_hint(__file__, request)],
    )
    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        pass
    return


@skipdeco
def test_oplev_iy(request, browser, plotsections, plot_verbosity):
    from wield.iirrational.file_io import load_any

    fname = matlab_test_data.matfiles["OplevPlant.mat"]
    fdict = load_any(fname=fname, ftype="mat")

    out = v2.data2filter(
        F_Hz=fdict["ap"]["iy"]["ff"],
        data=fdict["ap"]["iy"]["plant"],
        never_unstable_poles=True,
        never_unstable_zeros=True,
        mode="rational",
        hints=[stability_validate_and_plot_hint(__file__, request)],
    )
    print(out.fitter.poles)
    with plot_on_assert(__file__, request, out.fitter, plot_anyway=True):
        pass
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from wield.bunch import Bunch

from .base import (
    ArgumentError,
    # mapcheck_bool,
    # cplx_iIjJ,
)

data_args_hide = True


def mapcheck_cslist(aid, aname, val):
    if isinstance(val, (list, tuple)):
        return val

    vals = val.split(",")
    vals = [v for v in [v.strip() for v in vals] if v != ""]
    return vals


def mapcheck_choose(aid, aname, val):
    """
    How to decide the optimal order. Special parameters are:
     - "prompt", "shell", "interactive" (the default) To enter a command prompt that plots and requests
       a choice
     - "auto10" the default choice to use the shell if the baseline order is
       above 10 unless there is no choice. (10 may be changed)
     - "baseline" to use the baseline fit before the ChiSq starts rising.
     - "baseline10" to use the baseline "best fit" unless it is over order 10
                    (10 can be configured to other integers)
     - integer. To force a choice of order. It will always use the best fit
       under this order.
    """
    modes = ["shell", "prompt", "interactive", "auto10", "baseline", "baseline10", "best"]
    try:
        if isinstance(val, str):
            val = val.lower()
            if val in modes:
                return val
            elif val.startswith("baseline"):
                remainder = val[8:]
                if int(remainder) < 0:
                    raise ValueError()
                return val
            
        # this runs the valueerror test that it can be parsed
        if int(val) < 0:
            raise ValueError()
        return val
    except ValueError:
        raise ArgumentError(
            (
                "Argument {}={} must be an +integer or one of {}".format(
                    aname, val, modes
                )
            )
        )


def mapcheck_print_(aid, aname, val):
    """
    How to decide the optimal order. Special parameters are:
     - "prompt", "shell", "interactive" (the default) To enter a command prompt that plots and requests
       a choice
     - "auto10" the default choice to use the shell if the baseline order is
       above 10 unless there is no choice. (10 may be changed)
     - "baseline" to use the baseline fit before the ChiSq starts rising.
     - "baseline10" to use the baseline "best fit" unless it is over order 10
                    (10 can be configured to other integers)
     - integer. To force a choice of order. It will always use the best fit
       under this order.
    """
    modes = ["shell", "prompt", "interactive", "auto10", "baseline", "baseline10"]
    try:
        if isinstance(val, str):
            val = val.lower()
            if val in modes:
                return val
        # this runs the valueerror test that it can be parsed
        if int(val) < 0:
            raise ValueError()
        return val
    except ValueError:
        raise ArgumentError(
            (
                "Argument {}={} must be an +integer or one of {}".format(
                    aname, val, modes
                )
            )
        )



kw_hints = Bunch(
    datafile=dict(
        APpositional=True,
        APpriority=-2,
        APmetavar="data.file",
        about="""
        Specify data file. This file may be .h5, .hdf, .mat, .yaml.
        these filetypes have internal directory structure and the data are by
        default assumed to be in the parameters:
         'F_Hz', 'xfer', 'SNR', 'emphasis', inside of the 'data' group.
        Arguments such is -Xc and -F and -D in the "data" options group can override
        these keys. It is possible also to use csv files, but the syntax is complicated.
        """,
    ),
    outputfile=dict(
        APpositional=True,
        APpriority=-2,
        APmetavar="output.file",
        about="""
        Specify the output file to store fit results including zpk of the
        chosen and alternative fits, the configurations, and versioning information.
        Possible output extensions are .h5, .hdf, .mat, .pkl, .json, .yaml. Binary
        formats .mat, .h5, .pkl will include the original data, for full reconstruction
        of the fit. wield.iirrational may be called on the output file to rerun the fit.
        """,
    ),
    overwrite=dict(
        APaction="store_true",
        default=False,
        about="""
        Allow the output file to overwrite the previous output file (be careful).
        """,
    ),
    mode=dict(
        APshort="-m",
        APpriority=3,
        APchoices=[
            "full",
            "full2x",
            "fullAAA",
            "onlyAAA",
            "onlyAAAreduce",
            "AAA",
            "fit",
            "reduce",
            "rational",
            "rational2x",
            "dumpargs",
            "dumpargs_full",
            "printdata",
            "printconfig",
            "printsettings",
        ],
        default="full",
        about="""
        Fitting mode to use, to change the automation level. Must be one of
         - "full": to use rational fitting to get initial parameters, then alternate
                   optimization and order reduction
         - "full2x": runs the full optimization twice to refine it. Can often use a lower initial order.
         - "rational": to use rational fitting to get initial parameters and then
                       stop after the simplest order reduction. Useful to use with
                       other fitters. Does not perform delay fitting.
         - "fit": To take an initial ZPK guess, and only fit/optimize it. Good
                  for refining previous fits.
         - "reduce": To take an initial ZPK guess, and then alternate fitting and
                     order reduction.
         - "dumpargs": Dump arguments to determine current settings
         - "dumpargs_full": Dump arguments and all generated settings to see the full run setup
         - "printdata": Dump the layout of the data file specified
         - "printconfig": dump the layout of the config files specified
         - "printsettings": dump the fully loaded settings specified
        """,
    ),
    config=dict(
        APshort="-c",
        APaction="append",
        APpriority=0,
        default=[],
        about="""
        Specify configuration file. This file may be .json, .yaml, .ini, .h5, .mat.
        These will be interpreted as a dictionary or structure full of keyword
        arguments. This may be specified multiple times to overlay settings.
        Additional command line setting always override these configuration files.
        """,
    ),
    data_group=dict(
        APshort="-D",
        APgroup="groups",
        APpriority=1,
        default=["data", "IIRdata", "xfer"],
        mapcheck=mapcheck_cslist,
        about="""
        Group(s) within the loaded data file to search for data elements
        (see "data" section). May be a comma separated list to aggregate groups.
        defaults to "data,IIRdata,xfer". The first specified element will be
        the group that data is stored to the output file.
        """,
    ),
    config_group=dict(
        APgroup="groups",
        APpriority=1,
        default=["config", "conf", "IIRconfig", "IIRconf"],
        mapcheck=mapcheck_cslist,
        about="""
        Group(s) within the loaded config or data file to search for
        configuration elements. May be a comma separated list to aggregate groups.
        defaults to "config,conf,IIRconfig,IIRconf". The first element will be
        the group where configurations are stored to the output file.
        """,
    ),
    plot_fit=dict(
        APshort="-r",
        APgroup="plots",
        APpriority=10,
        default=None,
        about="""
        filename to plot (review) the chosen fit to. May be any extension supported by matplotlib.
        [.pdf, .svg, .png, .jpg, ...]
        """,
    ),
    plot_order=dict(
        APshort="-R",
        APgroup="plots",
        APpriority=10,
        default=None,
        about="""
        filename to plot (Review) potential orders and residuals to. May be any extension supported by matplotlib.
        [.pdf, .svg, .png, .jpg, ...]
        """,
    ),
    choose=dict(
        APshort="-j",
        APpriority=2,
        # TODO, type
        default="auto10",
        mapcheck=mapcheck_choose,
        about="""
        How to decide the optimal order. Special parameters are:
         - "prompt", "shell", "interactive" (the default) To enter a command prompt that plots and requests
           a choice
         - "auto10" the default choice to use the shell if the baseline order is
           above 10 unless there is no choice. (10 may be changed)
         - "baseline" to use the baseline fit before the ChiSq starts rising.
         - "baseline10" to use the baseline "best fit" unless it is over order 10
                        (10 can be configured to other integers)
         - integer. To force a choice of order. It will always use the best fit
           under this order.
        """,
    ),
    LIGO_foton=dict(
        APshort="-L",
        APpriority=2,
        APchoices=[
            "-",
            "none",
            "Sf",
        ],
        # TODO, type
        default="Sf",
        about="""
        What type of foton output would you prefer?
          - '-' or 'None' to suppress foton output
          - Sf for frequency domain (default)
          - n for normalized (not yet supported)
          - Sw for radial (not yet supported)
        """,
    ),
    refine=dict(
        APflags=["--no-refine"],
        APaction="store_false",
        default=True,
        APpriority=2,
        about="""
        disable the output hints of --zeros and --poles to run a second time and refine.
        """,
    ),
    refine_file=dict(
        APflags=["--refine_file"],
        default=None,
        APpriority=4,
        about="""
        indicate a file to write the hints of --zeros and --poles into.
        """,
    ),
    information=dict(
        APshort="-I",
        APpriority=6,
        default=None,
        about="""
        A string to detail what this fit is of. For documentation purposes.
        """,
    ),
    frequency=dict(
        APshort="-F",
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        mapcheck=mapcheck_cslist,
        default=[
            "frequency_Hz",
            "freq_Hz",
            "F_Hz",
            "f_Hz",
            "frequency",
            "freq",
            "F",
            "f",
        ],
        about="key or column number to find complex data array",
    ),
    dataXcomplex=dict(
        APshort="-Xc",
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        mapcheck=mapcheck_cslist,
        default=[
            "xfer",
            "data",
            "X",
            "transfer",
        ],
        about="key or column number to find complex data array",
    ),
    dataXreal=dict(
        APshort="-Xr",
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        mapcheck=mapcheck_cslist,
        default=[
            "xfer.real",
            "xfer.r",
            "xfer_real",
            "xfer_r",
            "data.real",
            "data.r",
            "data_real",
            "data_r",
            "X.real",
            "X.r",
            "Xreal",
            "Xr",
            "transfer_real",
            "transfer.real",
            "transfer_r",
            "transfer.r",
        ],
        about="key or column number to find data array",
    ),
    dataXimaginary=dict(
        APshort="-Xi",
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        mapcheck=mapcheck_cslist,
        default=[
            "xfer.imag",
            "xfer.i",
            "xfer_imag",
            "xfer_i",
            "X.imag",
            "X.i",
            "Ximag",
            "Xi",
            "transfer_imag",
            "transfer.imag",
            "transfer_i",
            "transfer.i",
        ],
        about="key or column number to find data array",
    ),
    dataXamplitude=dict(
        APflags=["-Xa", "-Xm", "--dataXamplitude", "--dataXmagnitude"],
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        mapcheck=mapcheck_cslist,
        default=[
            "xfer_mag",
            "xfer_magnitude",
            "xfer_amplitude",
            "xfer_amp" "Xa",
            "Xm",
            "Xamp",
            "Xmag" "amplitude",
            "magnitude",
            "amp",
            "mag",
        ],
        about="key or column number to find data array",
    ),
    dataXdB=dict(
        APflags=["-XdB", "-Xdb", "--dataXdB", "--dataXdb"],
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        default=["xfer_db", "xfer_dB", "Xdb", "XdB", "db", "dB"],
        about="key or column number to find data array",
    ),
    dataXphaseRad=dict(
        APshort="-Xp",
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        mapcheck=mapcheck_cslist,
        default=[
            "xfer_phase",
            "xfer_radians",
            "xfer_rad",
            "Xrad",
            "Xphase",
            "phase",
            "radians",
            "rad",
        ],
        about="key or column number to find data array",
    ),
    dataXphaseDeg=dict(
        APshort="-Xd",
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        mapcheck=mapcheck_cslist,
        default=[
            "xfer_degrees",
            "xfer_deg",
            "Xdeg",
            "Xdegrees",
            "degrees",
            "deg",
        ],
        about="key or column number to find data array",
    ),
    dataSNR=dict(
        APflags=["-W", "-S", "--dataSNR", "--dataWeights"],
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        mapcheck=mapcheck_cslist,
        default=[
            "SNR",
            "snr",
            "signal2noise" "weight",
            "Weight",
            "weights",
            "Weights",
            "W",
            "w",
        ],
        about="key or column number to find SNR array",
    ),
    dataCOH=dict(
        APflags=["-C", "--dataCOH"],
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        mapcheck=mapcheck_cslist,
        default=[
            "COH",
            "coh",
        ],
        about="""key or column numbers to find a coherence arrays. These are
        then used to compute the implied SNR. This argument can be specified
        in a comma-separated list to give multiple coherences to combine into
        a single SNR array.
        """,
    ),
    dataEmphasis=dict(
        APshort="-E",
        APhide=data_args_hide,
        APgroup="data",
        APpriority=5,
        APmetavar="key",
        default=["emphasis", "emph", "Emph", "E", "e"],
        mapcheck=mapcheck_cslist,
        about="key or column number to find data array",
    ),
)

groups_kw = dict(
    groups=dict(
        APpriority=4,
        about="""
        The data and config file formats (except csv) all load internally to a
        dictionary or structure representation. These arguments specify the
        keys or groups storing the relevant data or configuration subkeys. These
        also control the output groups when saving files.
    """,
    ),
    data=dict(
        APpriority=5,
        about="""
    Specify keys to search within the data group for specific arrays.
        -F,   --frequency,
        -Xc,  --dataXcomplex
        -Xr,  --dataXreal,      -Xi,  --dataXimaginary
        -Xa,  --dataXamplitude, -Xm,  --dataXmagnitude
        -XdB, --dataXdB,        -Xdb, --dataXdb
        -Xd,  --dataXphaseDeg,  -Xp,  --dataXphaseRad
        -S,   --dataSNR,        -W,   --dataW
        -E,   --dataEmphasis
    """,
    ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from wield.bunch import Bunch
from .utilities import args
import scipy.io


def data2testcase(
    argB=args.UNSPEC,
    data=args.UNSPEC,
    F_Hz=args.UNSPEC,
    SNR=args.UNSPEC,
    fname=args.UNSPEC,
    F_nyquist_Hz=args.UNSPEC,
    bestfit_ZPK_z=args.UNSPEC,
    bestfit_ZPK_s=args.UNSPEC,
    description=args.UNSPEC,
):
    """
    Save a testcase in in matlab .mat format with a standardized .dat for standard read testcase reading.
    :arg: data
    """
    data = args.argscan(locals(), argB, args.REQ, arg="data")
    F_Hz = args.argscan(locals(), argB, args.REQ, arg="F_Hz")
    SNR = args.argscan(locals(), argB, 1, arg="SNR")
    fname = args.argscan(locals(), argB, args.REQ, arg="fname")
    F_nyquist_Hz = args.argscan(locals(), argB, args.REQ, arg="F_nyquist_Hz")
    bestfit_ZPK_z = args.argscan(locals(), argB, None, arg="bestfit_ZPK_z")
    bestfit_ZPK_s = args.argscan(locals(), argB, None, arg="bestfit_ZPK_s")
    description = args.argscan(locals(), argB, args.REQ, arg="description")

    mdict = dict(
        data=data,
        F_Hz=F_Hz,
        SNR=SNR,
        description=description,
        version=1,
    )

    if bestfit_ZPK_s is not None:
        mdict["bestfit_ZPK_s"] = bestfit_ZPK_s

    if bestfit_ZPK_z is not None:
        mdict["bestfit_ZPK_z"] = bestfit_ZPK_z

    return scipy.io.savemat(fname, mdict)


def testcase2data(fname):
    import scipy.io

    data = scipy.io.loadmat(fname, squeeze_me=True)
    return Bunch(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


def fitter_order(fitter):
    return max(len(fitter.zeros), len(fitter.poles))


class FitAidFake(object):
    def __init__(
        self,
    ):
        return

    def hint(self, *args, **kwargs):
        default = kwargs.get("default")
        return default

    def hint_arg(self, func_arg, *args, **kwargs):
        if func_arg is None:
            return self.hint(*args, **kwargs)
        else:
            return func_arg

    def log(self, *args, **kwargs):
        print(*args)

    def log_info(self, *args, **kwargs):
        print(*args)

    def log_warn(self, *args, **kwargs):
        print(*args)

    def log_alert(self, *args, **kwargs):
        print(*args)

    def log_debug(self, *args, **kwargs):
        print(*args)

    def log_rationale(self, *args, **kwargs):
        print(*args)


def ensure_aid(aid=None, **kwargs):
    if aid is None:
        return FitAidFake()
    return aid



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from wield.bunch import Bunch

# import numpy as np
# import scipy.linalg

from .base import CodingType, Ipi, I2pi

from ..MRF import MultiReprFilterS
from .gain_delay import CodingGainDelay
from .cplx_real_imag import CodingRI
from .cplx_frequency_BW import CodingFBW
from .cplx_nl_freq_nl_BW import CodingnlFnlBW

# from .cplx_nl_freq_amplitude import CodingnlFA
from .real_BW import CodingRealBW
from .real_nl_BW import CodingRealnlBW
from .cplx_sos import CodingSOS, CodingSOSMirror

# from .multi_order_section import CodingMOS

import sys

# get the current module object to provide a reference
_module = sys.modules[__name__]


def is_unstable_S(root):
    return root.real > 0


_common = dict(
    module=_module,
    representation="Sf",
    mrf_default=MultiReprFilterS,
    is_unstable=is_unstable_S,
)

coding_maps = Bunch(
    RI=Bunch(
        gain_delay=CodingGainDelay,
        num_r=CodingRealBW,
        num_c=CodingRI,
        den_r=CodingRealBW,
        den_c=CodingRI,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    FBW=Bunch(
        gain_delay=CodingGainDelay,
        num_r=CodingRealBW,
        num_c=CodingFBW,
        den_r=CodingRealBW,
        den_c=CodingFBW,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    nlFBW=Bunch(
        gain_delay=CodingGainDelay,
        num_r=CodingRealnlBW,
        num_c=CodingnlFnlBW,
        den_r=CodingRealnlBW,
        den_c=CodingnlFnlBW,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    nlFBW_safe=Bunch(
        gain_delay=CodingGainDelay,
        num_r=CodingRealnlBW,
        num_c=CodingnlFnlBW,
        den_r=CodingRealnlBW,
        den_c=CodingnlFnlBW,
        num_r_u=CodingRealBW,
        num_c_u=CodingFBW,
        den_r_u=CodingRealBW,
        den_c_u=CodingFBW,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    # TODO, not ready, gotta be able to deal with roots at 0
    SOS=Bunch(
        gain_delay=CodingGainDelay,
        num_r=CodingSOS,
        num_c=CodingSOS,
        den_r=CodingSOS,
        den_c=CodingSOS,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=2,
        den_collect_r=2,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    # SOS_safe = Bunch(
    #    gain_delay    = CodingGainDelay,
    #    num_r         = CodingSOSMirror,
    #    num_c         = CodingSOSMirror,
    #    den_r         = CodingSOSMirror,
    #    den_c         = CodingSOSMirror,
    #    num_r_u       = CodingSOS,
    #    num_c_u       = CodingSOS,
    #    den_r_u       = CodingSOS,
    #    den_c_u       = CodingSOS,
    #    num_collect_r = 2,
    #    den_collect_r = 2,
    #    num_collect_c = 1,
    #    den_collect_c = 1,
    #    **_common
    # ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.cm as cmx
import matplotlib.colors as colors


import contextlib
from matplotlib.ticker import LogLocator, AutoLocator
from matplotlib.scale import LogScale
from matplotlib import transforms as mtransforms
import math


@contextlib.contextmanager
def savefig(save_name, size_x_in, size_y_in):
    fig = plt.figure()
    yield fig
    fig.set_size_inches(size_x_in, size_y_in)
    fig.savefig(save_name)


def hsva_to_rgba(h, s, v, a=1.0):
    """
    Convert h,s,v into rgb using mpl. The converter of mpl is vectorized, so it is not convenient for
    this operation, this is convenient. Furthermore, this will correctly wrap the h value
    """

    s = min(s, 1.0)
    s = max(s, 0.0)
    v = min(v, 1.0)
    v = max(v, 0.0)
    mpl_formatted = np.array([[[h % 1.0, s, v]]])
    rgbarray = mpl.colors.hsv_to_rgb(mpl_formatted)[0, 0]
    return np.concatenate([rgbarray, [a]])


def vertical_stack_optional(ax_list, ax_names, fig=None, autofig=True, share_x=False):
    if autofig and all(ax is None for ax in ax_list):
        import matplotlib.pylab as plt

        if fig is None:
            fig = plt.figure()
        for idx in range(len(ax_list)):
            if idx > 0 or (not share_x):
                ax = fig.add_subplot(len(ax_list), 1, idx + 1)
            else:
                ax = fig.add_subplot(len(ax_list), 1, idx + 1, sharex=ax_list[0])
            ax_list[idx] = ax
    ax_dict = DictTree()
    for name, ax in zip(ax_names, ax_list):
        ax_dict.local_set(name, ax)
    return ax_dict


class LogLocator2(LogLocator):
    """
    Determine the tick locations for log axes
    """

    def view_limits(self, vmin, vmax):
        "Try to choose the view limits intelligently"
        b = self._base

        if vmax < vmin:
            vmin, vmax = vmax, vmin

        if self.axis.axes.name == "polar":
            vmax = math.ceil(math.log(vmax) / math.log(b))
            vmin = b ** (vmax - self.numdecs)
            return vmin, vmax

        minpos = self.axis.get_minpos()

        if minpos <= 0 or not np.isfinite(minpos):
            raise ValueError(
                "Data has no positive values, and therefore can not be " "log-scaled."
            )

        if vmin <= minpos:
            vmin = minpos
        result = mtransforms.nonsingular(vmin, vmax)
        return result


class AutoLocator2(AutoLocator):
    """ """

    def tick_values(self, vmin, vmax):
        locs = super(AutoLocator2, self).tick_values(vmin, vmax)
        locs = np.asarray(locs)
        tidx = np.searchsorted(locs, vmax)
        if tidx == len(locs):
            tidx -= 1
        ticksep = locs[tidx] - locs[tidx - 1]
        tickmax = vmax - ticksep / 2
        locs = locs[locs < tickmax]
        return locs


class LogScale2(LogScale):
    name = "log_zoom"

    def set_default_locators_and_formatters(self, axis):
        """
        Set the locators and formatters to specialized versions for
        log scaling.
        """
        super(LogScale2, self).set_default_locators_and_formatters(axis)
        axis.set_major_locator(LogLocator2(self.base))
        axis.set_minor_locator(LogLocator2(self.base, self.subs))


mpl.scale.register_scale(LogScale2)


def indexed_cmap(N, cmap="hsv", offset=0):
    """Returns a function that maps each index in 0, 1, ... N-1 to a distinct
    RGB color."""
    color_norm = colors.Normalize(vmin=0, vmax=N)
    scalar_map = cmx.ScalarMappable(norm=color_norm, cmap=cmap)

    def map_index_to_rgb_color(index):
        return scalar_map.to_rgba(index + offset)

    return map_index_to_rgb_color
